from typing import Optional

from sqlalchemy import Column, types
from sqlalchemy.orm.exc import NoResultFound

from .base import Model


class Supportparams(Model):
    supportparams_id = Column(types.Integer, primary_key=True)
    document_path = Column(types.String(255), nullable=False, unique=True)
    systempid = Column(types.String(255))
    systemos = Column(types.String(255))
    systemperlv = Column(types.String(255))
    systemperlexe = Column(types.String(255))
    idstring = Column(types.String(255))
    program = Column(types.String(255))
    commandline = Column(types.Text)
    sampleconfig_path = Column(types.String(255))
    sampleconfig = Column(types.Text(16777215))
    time = Column(types.DateTime)

    def __repr__(self):
        return "{self.__class__.__name__}: {self.document_path}".format(self=self)



"""Code that handles CLI commands to upload"""
import datetime as dt
import logging
from pathlib import Path
from typing import Set

import click

from cgmodels.trailblazer.constants import AnalysisTypes
from cg.apps.tb import TrailblazerAPI
from cg.constants import Pipeline
from cg.constants.constants import DRY_RUN
from cg.constants.delivery import PIPELINE_ANALYSIS_TAG_MAP
from cg.constants.priority import PRIORITY_TO_SLURM_QOS
from cg.meta.deliver import DeliverAPI
from cg.meta.rsync import RsyncAPI
from cg.store import Store
from cg.store.models import Family

LOG = logging.getLogger(__name__)


@click.command("clinical-delivery")
@click.pass_context
@click.argument("case_id", required=True)
@DRY_RUN
def upload_clinical_delivery(context: click.Context, case_id: str, dry_run: bool):
    """Links the appropriate files for a case, based on the data_delivery, to the customer folder
    and subsequently uses rsync to upload it to caesar."""

    click.echo(click.style("----------------- Clinical-delivery -----------------"))

    case_obj: Family = context.obj.status_db.get_case_by_internal_id(internal_id=case_id)
    delivery_types: Set[str] = case_obj.get_delivery_arguments()
    is_sample_delivery: bool
    is_case_delivery: bool
    is_complete_delivery: bool
    job_id: int
    is_sample_delivery, is_case_delivery = DeliverAPI.get_delivery_scope(
        delivery_arguments=delivery_types
    )
    if not delivery_types:
        LOG.info(f"No delivery of files requested for case {case_id}")
        return

    LOG.debug("Delivery types are: %s", delivery_types)
    for delivery_type in delivery_types:
        DeliverAPI(
            store=context.obj.status_db,
            hk_api=context.obj.housekeeper_api,
            case_tags=PIPELINE_ANALYSIS_TAG_MAP[delivery_type]["case_tags"],
            sample_tags=PIPELINE_ANALYSIS_TAG_MAP[delivery_type]["sample_tags"],
            delivery_type=delivery_type,
            project_base_path=Path(context.obj.delivery_path),
        ).deliver_files(case_obj=case_obj)

    rsync_api: RsyncAPI = RsyncAPI(context.obj)
    is_complete_delivery, job_id = rsync_api.slurm_rsync_single_case(
        case_id=case_id,
        dry_run=dry_run,
        sample_files_present=is_sample_delivery,
        case_files_present=is_case_delivery,
    )
    RsyncAPI.write_trailblazer_config(
        {"jobs": [str(job_id)]}, config_path=rsync_api.trailblazer_config_path
    )
    analysis_name: str = f"{case_id}_rsync" if is_complete_delivery else f"{case_id}_partial"
    if not dry_run:
        context.obj.trailblazer_api.add_pending_analysis(
            case_id=analysis_name,
            analysis_type=AnalysisTypes.OTHER,
            config_path=rsync_api.trailblazer_config_path.as_posix(),
            out_dir=rsync_api.log_dir.as_posix(),
            slurm_quality_of_service=PRIORITY_TO_SLURM_QOS[case_obj.priority],
            data_analysis=Pipeline.RSYNC,
            ticket=case_obj.latest_ticket,
        )
    LOG.info("Transfer of case %s started with SLURM job id %s", case_id, job_id)


@click.command("all-fastq")
@click.pass_context
@DRY_RUN
def auto_fastq(context: click.Context, dry_run: bool):
    """Starts upload of all not previously uploaded cases with analysis type fastq to
    clinical-delivery."""

    status_db: Store = context.obj.status_db
    trailblazer_api: TrailblazerAPI = context.obj.trailblazer_api
    for analysis_obj in status_db.get_analyses_to_upload(pipeline=Pipeline.FASTQ):
        if analysis_obj.family.analyses[0].uploaded_at:
            LOG.warning(
                "Newer analysis already uploaded for %s, skipping",
                analysis_obj.family.internal_id,
            )
            continue
        if analysis_obj.upload_started_at:
            if trailblazer_api.is_latest_analysis_completed(
                case_id=analysis_obj.family.internal_id
            ):
                LOG.info(
                    "The upload for %s is completed, setting uploaded at to %s",
                    analysis_obj.family.internal_id,
                    dt.datetime.now(),
                )
                analysis_obj.uploaded_at = dt.datetime.now()
            else:
                LOG.warning(
                    "Upload to clinical-delivery for %s has already started, skipping",
                    analysis_obj.family.internal_id,
                )
            continue
        case: Family = analysis_obj.family
        LOG.info("Uploading family: %s", case.internal_id)
        analysis_obj.upload_started_at = dt.datetime.now()
        context.invoke(upload_clinical_delivery, case_id=case.internal_id, dry_run=dry_run)
        if not dry_run:
            status_db.session.commit()



import datetime as dt
import logging
import os
import shutil
from pathlib import Path
from subprocess import CalledProcessError
from typing import List, Optional, Tuple, Union

import click
from housekeeper.store.models import Bundle, Version

from cg.apps.environ import environ_email
from cg.constants import CASE_ACTIONS, EXIT_FAIL, EXIT_SUCCESS, Pipeline, Priority
from cg.constants.constants import AnalysisType, WorkflowManager
from cg.constants.priority import PRIORITY_TO_SLURM_QOS
from cg.exc import BundleAlreadyAddedError, CgDataError, CgError
from cg.meta.meta import MetaAPI
from cg.meta.workflow.fastq import FastqHandler
from cg.models.analysis import AnalysisModel
from cg.models.cg_config import CGConfig
from cg.store.models import Analysis, BedVersion, Family, FamilySample, Sample

LOG = logging.getLogger(__name__)


class AnalysisAPI(MetaAPI):
    """
    Parent class containing all methods that are either shared or overridden by other workflow APIs
    """

    def __init__(self, pipeline: Pipeline, config: CGConfig):
        super().__init__(config=config)
        self.pipeline = pipeline
        self._process = None

    @property
    def root(self):
        raise NotImplementedError

    @property
    def threshold_reads(self):
        """Defines whether the threshold for adequate read count should be passed for all samples
        when determining if the analysis for a case should be automatically started"""
        return False

    @property
    def process(self):
        raise NotImplementedError

    @property
    def fastq_handler(self):
        return FastqHandler

    @staticmethod
    def get_help(context):
        """
        If no argument is passed, print help text
        """
        if context.invoked_subcommand is None:
            click.echo(context.get_help())

    def verify_deliverables_file_exists(self, case_id: str) -> None:
        if not Path(self.get_deliverables_file_path(case_id=case_id)).exists():
            raise CgError(f"No deliverables file found for case {case_id}")

    def verify_case_config_file_exists(self, case_id: str) -> None:
        if not Path(self.get_case_config_path(case_id=case_id)).exists():
            raise CgError(f"No config file found for case {case_id}")

    def verify_case_id_in_statusdb(self, case_id: str) -> None:
        """Passes silently if case exists in StatusDB, raises error if case is missing"""

        case_obj: Family = self.status_db.get_case_by_internal_id(internal_id=case_id)
        if not case_obj:
            LOG.error("Case %s could not be found in StatusDB!", case_id)
            raise CgError
        if not case_obj.links:
            LOG.error("Case %s has no samples in in StatusDB!", case_id)
            raise CgError
        LOG.info("Case %s exists in status db", case_id)

    def check_analysis_ongoing(self, case_id: str) -> None:
        if self.trailblazer_api.is_latest_analysis_ongoing(case_id=case_id):
            LOG.warning(f"{case_id} : analysis is still ongoing - skipping")
            raise CgError(f"Analysis still ongoing in Trailblazer for case {case_id}")

    def verify_case_path_exists(self, case_id: str) -> None:
        """Check if case path exists."""
        if not self.get_case_path(case_id=case_id).exists():
            LOG.info(f"No working directory for {case_id} exists")
            raise FileNotFoundError(f"No working directory for {case_id} exists")

    def get_priority_for_case(self, case_id: str) -> int:
        """Get priority from the status db case priority"""
        case_obj: Family = self.status_db.get_case_by_internal_id(internal_id=case_id)
        return case_obj.priority.value or Priority.research

    def get_slurm_qos_for_case(self, case_id: str) -> str:
        """Get Quality of service (SLURM QOS) for the case."""
        priority: int = self.get_priority_for_case(case_id=case_id)
        return PRIORITY_TO_SLURM_QOS[priority]

    def get_workflow_manager(self) -> str:
        """Get workflow manager for a given pipeline."""
        return WorkflowManager.Slurm.value

    def get_case_path(self, case_id: str) -> Union[List[Path], Path]:
        """Path to case working directory."""
        raise NotImplementedError

    def get_case_config_path(self, case_id) -> Path:
        """Path to case config file"""
        raise NotImplementedError

    def get_trailblazer_config_path(self, case_id: str) -> Path:
        """Path to Trailblazer job id file"""
        raise NotImplementedError

    def get_sample_name_from_lims_id(self, lims_id: str) -> str:
        """Retrieve sample name provided by customer for specific sample"""
        sample: Sample = self.status_db.get_sample_by_internal_id(internal_id=lims_id)
        return sample.name

    def link_fastq_files(self, case_id: str, dry_run: bool = False) -> None:
        """
        Links fastq files from Housekeeper to case working directory
        """
        raise NotImplementedError

    def get_bundle_deliverables_type(self, case_id: str) -> Optional[str]:
        return None

    @staticmethod
    def get_application_type(sample_obj: Sample) -> str:
        """
        Gets application type for sample. Only application types supported by trailblazer (or other)
        are valid outputs.
        """
        prep_category: str = sample_obj.prep_category
        if prep_category and prep_category.lower() in {
            AnalysisType.TARGETED_GENOME_SEQUENCING,
            AnalysisType.WHOLE_EXOME_SEQUENCING,
            AnalysisType.WHOLE_GENOME_SEQUENCING,
            AnalysisType.WHOLE_TRANSCRIPTOME_SEQUENCING,
        }:
            return prep_category.lower()
        return AnalysisType.OTHER

    def upload_bundle_housekeeper(self, case_id: str, dry_run: bool = False) -> None:
        """Storing bundle data in Housekeeper for CASE_ID"""
        LOG.info(f"Storing bundle data in Housekeeper for {case_id}")
        bundle_data: dict = self.get_hermes_transformed_deliverables(case_id)
        bundle_result: Tuple[Bundle, Version] = self.housekeeper_api.add_bundle(
            bundle_data=bundle_data
        )
        if not bundle_result:
            LOG.info("Bundle already added to Housekeeper!")
            raise BundleAlreadyAddedError("Bundle already added to Housekeeper!")
        bundle_object, bundle_version = bundle_result
        if dry_run:
            LOG.info("Dry-run: Housekeeper changes will not be commited")
            LOG.info(
                "The following files would be stored:\n%s",
                "\n".join([f["path"] for f in bundle_data["files"]]),
            )
            return
        self.housekeeper_api.include(bundle_version)
        self.housekeeper_api.add_commit(bundle_object)
        self.housekeeper_api.add_commit(bundle_version)
        LOG.info(
            f"Analysis successfully stored in Housekeeper: {case_id} : {bundle_version.created_at}"
        )

    def upload_bundle_statusdb(self, case_id: str, dry_run: bool = False) -> None:
        """Storing analysis bundle in StatusDB for CASE_ID"""

        LOG.info(f"Storing analysis in StatusDB for {case_id}")
        case_obj: Family = self.status_db.get_case_by_internal_id(internal_id=case_id)
        analysis_start: dt.datetime = self.get_bundle_created_date(case_id=case_id)
        pipeline_version: str = self.get_pipeline_version(case_id=case_id)
        new_analysis: Family = self.status_db.add_analysis(
            pipeline=self.pipeline,
            version=pipeline_version,
            started_at=analysis_start,
            completed_at=dt.datetime.now(),
            primary=(len(case_obj.analyses) == 0),
        )
        new_analysis.family = case_obj
        if dry_run:
            LOG.info("Dry-run: StatusDB changes will not be commited")
            return
        self.status_db.session.add(new_analysis)
        self.status_db.session.commit()
        LOG.info(f"Analysis successfully stored in StatusDB: {case_id} : {analysis_start}")

    def get_deliverables_file_path(self, case_id: str) -> Path:
        raise NotImplementedError

    def get_analysis_finish_path(self, case_id: str) -> Path:
        raise NotImplementedError

    def add_pending_trailblazer_analysis(self, case_id: str) -> None:
        self.check_analysis_ongoing(case_id=case_id)
        self.trailblazer_api.mark_analyses_deleted(case_id=case_id)
        self.trailblazer_api.add_pending_analysis(
            case_id=case_id,
            email=environ_email(),
            analysis_type=self.get_application_type(
                self.status_db.get_case_by_internal_id(internal_id=case_id).links[0].sample
            ),
            out_dir=self.get_trailblazer_config_path(case_id=case_id).parent.as_posix(),
            config_path=self.get_trailblazer_config_path(case_id=case_id).as_posix(),
            slurm_quality_of_service=self.get_slurm_qos_for_case(case_id=case_id),
            data_analysis=str(self.pipeline),
            ticket=self.status_db.get_latest_ticket_from_case(case_id),
            workflow_manager=self.get_workflow_manager(),
        )

    def get_hermes_transformed_deliverables(self, case_id: str) -> dict:
        return self.hermes_api.create_housekeeper_bundle(
            bundle_name=case_id,
            deliverables=self.get_deliverables_file_path(case_id=case_id),
            pipeline=str(self.pipeline),
            analysis_type=self.get_bundle_deliverables_type(case_id),
            created=self.get_bundle_created_date(case_id),
        ).dict()

    def get_bundle_created_date(self, case_id: str) -> dt.datetime:
        return self.get_date_from_file_path(self.get_deliverables_file_path(case_id=case_id))

    def get_pipeline_version(self, case_id: str) -> str:
        """
        Calls the pipeline to get the pipeline version number. If fails, returns a placeholder value instead.
        """
        try:
            self.process.run_command(["--version"])
            return list(self.process.stdout_lines())[0].split()[-1]
        except (Exception, CalledProcessError):
            LOG.warning("Could not retrieve %s workflow version!", self.pipeline)
            return "0.0.0"

    def set_statusdb_action(
        self, case_id: str, action: Optional[str], dry_run: bool = False
    ) -> None:
        """
        Set one of the allowed actions on a case in StatusDB.
        """
        if dry_run:
            LOG.info(f"Dry-run: Action {action} would be set for case {case_id}")
            return
        if action in [None, *CASE_ACTIONS]:
            case_obj: Family = self.status_db.get_case_by_internal_id(internal_id=case_id)
            case_obj.action = action
            self.status_db.session.commit()
            LOG.info("Action %s set for case %s", action, case_id)
            return
        LOG.warning(
            f"Action '{action}' not permitted by StatusDB and will not be set for case {case_id}"
        )

    def get_analyses_to_clean(self, before: dt.datetime) -> List[Analysis]:
        analyses_to_clean = self.status_db.get_analyses_to_clean(
            pipeline=self.pipeline, before=before
        )
        return analyses_to_clean

    def get_cases_to_analyze(self) -> List[Family]:
        return self.status_db.cases_to_analyze(
            pipeline=self.pipeline, threshold=self.threshold_reads
        )

    def get_running_cases(self) -> List[Family]:
        return self.status_db.get_running_cases_in_pipeline(pipeline=self.pipeline)

    def get_cases_to_store(self) -> List[Family]:
        """Retrieve a list of cases where analysis finished successfully,
        and is ready to be stored in Housekeeper"""
        return [
            case_object
            for case_object in self.get_running_cases()
            if self.trailblazer_api.is_latest_analysis_completed(case_id=case_object.internal_id)
        ]

    def get_sample_fastq_destination_dir(self, case: Family, sample: Sample):
        """Return the path to the FASTQ destination directory."""
        raise NotImplementedError

    def gather_file_metadata_for_sample(self, sample_obj: Sample) -> List[dict]:
        return [
            self.fastq_handler.parse_file_data(file_obj.full_path)
            for file_obj in self.housekeeper_api.files(
                bundle=sample_obj.internal_id, tags=["fastq"]
            )
        ]

    def link_fastq_files_for_sample(
        self, case_obj: Family, sample_obj: Sample, concatenate: bool = False
    ) -> None:
        """
        Link FASTQ files for a sample to working directory.
        If pipeline input requires concatenated fastq, files can also be concatenated
        """
        linked_reads_paths = {1: [], 2: []}
        concatenated_paths = {1: "", 2: ""}
        files: List[dict] = self.gather_file_metadata_for_sample(sample_obj=sample_obj)
        sorted_files = sorted(files, key=lambda k: k["path"])
        fastq_dir = self.get_sample_fastq_destination_dir(case=case_obj, sample=sample_obj)
        fastq_dir.mkdir(parents=True, exist_ok=True)

        for fastq_data in sorted_files:
            fastq_path = Path(fastq_data["path"])
            fastq_name = self.fastq_handler.create_fastq_name(
                lane=fastq_data["lane"],
                flowcell=fastq_data["flowcell"],
                sample=sample_obj.internal_id,
                read=fastq_data["read"],
                undetermined=fastq_data["undetermined"],
                meta=self.get_additional_naming_metadata(sample_obj),
            )
            destination_path: Path = fastq_dir / fastq_name
            linked_reads_paths[fastq_data["read"]].append(destination_path)
            concatenated_paths[
                fastq_data["read"]
            ] = f"{fastq_dir}/{self.fastq_handler.get_concatenated_name(fastq_name)}"

            if not destination_path.exists():
                LOG.info(f"Linking: {fastq_path} -> {destination_path}")
                destination_path.symlink_to(fastq_path)
            else:
                LOG.warning(f"Destination path already exists: {destination_path}")

        if not concatenate:
            return

        LOG.info("Concatenation in progress for sample %s.", sample_obj.internal_id)
        for read, value in linked_reads_paths.items():
            self.fastq_handler.concatenate(linked_reads_paths[read], concatenated_paths[read])
            self.fastq_handler.remove_files(value)

    def get_target_bed_from_lims(self, case_id: str) -> Optional[str]:
        """Get target bed filename from LIMS."""
        case: Family = self.status_db.get_case_by_internal_id(internal_id=case_id)
        sample: Sample = case.links[0].sample
        if sample.from_sample:
            sample: Sample = self.status_db.get_sample_by_internal_id(
                internal_id=sample.from_sample
            )
        target_bed_shortname: str = self.lims_api.capture_kit(lims_id=sample.internal_id)
        if not target_bed_shortname:
            return None
        bed_version: Optional[BedVersion] = self.status_db.get_bed_version_by_short_name(
            bed_version_short_name=target_bed_shortname
        )
        if not bed_version:
            raise CgDataError(f"Bed-version {target_bed_shortname} does not exist")
        return bed_version.filename

    def decompression_running(self, case_id: str) -> None:
        """Check if decompression is running for a case"""
        is_decompression_running: bool = self.prepare_fastq_api.is_spring_decompression_running(
            case_id
        )
        if not is_decompression_running:
            LOG.warning(
                "Decompression can not be started for %s",
                case_id,
            )
            return
        LOG.info(
            "Decompression is running for %s, analysis will be started when decompression is done",
            case_id,
        )
        self.set_statusdb_action(case_id=case_id, action="analyze")

    def decompress_case(self, case_id: str, dry_run: bool) -> None:
        """Decompress all possible fastq files for all samples in a case"""
        LOG.info(
            "The analysis for %s could not start, decompression is needed",
            case_id,
        )
        decompression_possible: bool = (
            self.prepare_fastq_api.can_at_least_one_sample_be_decompressed(case_id)
        )
        if not decompression_possible:
            self.decompression_running(case_id=case_id)
            return
        case_obj: Family = self.status_db.get_case_by_internal_id(internal_id=case_id)
        link: FamilySample
        any_decompression_started = False
        for link in case_obj.links:
            sample_id: str = link.sample.internal_id
            if dry_run:
                LOG.info(
                    f"This is a dry run, therefore decompression for {sample_id} won't be started"
                )
                continue
            decompression_started: bool = self.prepare_fastq_api.compress_api.decompress_spring(
                sample_id
            )
            if decompression_started:
                any_decompression_started = True

        if not any_decompression_started:
            LOG.warning("Decompression failed to start for %s", case_id)
            return
        if not dry_run:
            self.set_statusdb_action(case_id=case_id, action="analyze")
        LOG.info("Decompression started for %s", case_id)

    def resolve_decompression(self, case_id: str, dry_run: bool) -> bool:
        """
        Handles decompression automatically for case.
        Return True if decompression is running
        """
        if self.prepare_fastq_api.is_spring_decompression_needed(case_id):
            self.decompress_case(case_id=case_id, dry_run=dry_run)
            return True

        if self.prepare_fastq_api.is_spring_decompression_running(case_id):
            self.set_statusdb_action(case_id=case_id, action="analyze")
            return True

        LOG.info("Linking fastq files in housekeeper for case %s", case_id)
        self.prepare_fastq_api.check_fastq_links(case_id)

        if self.prepare_fastq_api.is_spring_decompression_needed(case_id):
            return True

        LOG.info("Decompression for case %s not needed", case_id)
        return False

    @staticmethod
    def get_date_from_file_path(file_path: Path) -> dt.datetime.date:
        """
        Get date from deliverables path using date created metadata.
        """
        return dt.datetime.fromtimestamp(int(os.path.getctime(file_path)))

    def get_additional_naming_metadata(self, sample_obj: Sample) -> Optional[str]:
        return None

    def get_latest_metadata(self, case_id: str) -> AnalysisModel:
        """Get the latest metadata of a specific case"""

        raise NotImplementedError

    def parse_analysis(
        self, config_raw: dict, qc_metrics_raw: dict, sample_info_raw: dict
    ) -> AnalysisModel:
        """Parses output analysis files"""

        raise NotImplementedError

    def clean_analyses(self, case_id: str) -> None:
        """Add a cleaned at date for all analyses related to a case."""
        analyses: list = self.status_db.get_case_by_internal_id(internal_id=case_id).analyses
        LOG.info(f"Adding a cleaned at date for case {case_id}")
        for analysis_obj in analyses:
            analysis_obj.cleaned_at = analysis_obj.cleaned_at or dt.datetime.now()
            self.status_db.session.commit()

    def clean_run_dir(self, case_id: str, yes: bool, case_path: Union[List[Path], Path]) -> int:
        """Remove workflow run directory."""

        try:
            self.verify_case_path_exists(case_id=case_id)
        except FileNotFoundError:
            self.clean_analyses(case_id)

        if yes or click.confirm(f"Are you sure you want to remove all files in {case_path}?"):
            if case_path.is_symlink():
                LOG.warning(
                    f"Will not automatically delete symlink: {case_path}, delete it manually",
                )
                return EXIT_FAIL

            shutil.rmtree(case_path, ignore_errors=True)
            LOG.info(f"Cleaned {case_path}")
            self.clean_analyses(case_id=case_id)
            return EXIT_SUCCESS



import logging
import shutil
from pathlib import Path

import pytest

from cg.apps.cgstats.stats import StatsAPI
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.flow_cell import FlowCell
from cg.utils import Process
from tests.apps.cgstats.conftest import fixture_populated_stats_api, fixture_stats_api
from tests.apps.demultiplex.conftest import (
    fixture_demux_run_dir_bcl2fastq,
    fixture_demux_run_dir_dragen,
    fixture_lims_novaseq_bcl2fastq_samples,
    fixture_lims_novaseq_dragen_samples,
    fixture_lims_novaseq_samples,
    fixture_novaseq_dir_bcl2fastq,
    fixture_novaseq_dir_dragen,
)

LOG = logging.getLogger(__name__)


@pytest.fixture(name="demux_results_not_finished_dir")
def fixture_demux_results_not_finished_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to a dir with demultiplexing results where demux has been done but nothing is cleaned."""
    return Path(demultiplex_fixtures, "demultiplexed-runs-unfinished")


@pytest.fixture(name="novaseq_bcl2fastq_sample_sheet_path")
def fixture_novaseq_bcl2fastq_sample_sheet_path(demultiplex_fixtures: Path) -> Path:
    """Return the path to a Novaseq bcl2fastq sample sheet."""
    return Path(demultiplex_fixtures, "SampleSheetS2_Bcl2Fastq.csv")


@pytest.fixture(name="flow_cell_runs_working_directory")
def fixture_flow_cell_runs_working_directory(project_dir: Path) -> Path:
    """Return the path to a working directory with flow cells ready for demux."""
    working_dir: Path = Path(project_dir, "flowcell-runs")
    working_dir.mkdir(parents=True)
    return working_dir


@pytest.fixture(name="flow_cell_runs_working_directory_bcl2fastq")
def fixture_flow_cell_runs_working_directory_bcl2fastq(
    flow_cell_runs_working_directory: Path,
) -> Path:
    """Return the path to a working directory with flow cells ready for demux."""
    working_dir: Path = Path(flow_cell_runs_working_directory, "bcl2fastq")
    working_dir.mkdir(parents=True)
    return working_dir


@pytest.fixture(name="flow_cell_runs_working_directory_dragen")
def fixture_flow_cell_runs_working_directory_dragen(flow_cell_runs_working_directory: Path) -> Path:
    """Return the path to a working directory with flow cells ready for demux."""
    working_dir: Path = Path(flow_cell_runs_working_directory, "dragen")
    working_dir.mkdir(parents=True)
    return working_dir


@pytest.fixture(name="demultiplexed_flow_cells_working_directory")
def fixture_demultiplexed_flow_cells_working_directory(project_dir: Path) -> Path:
    """Return the path to a working directory with flow cells that have been demultiplexed."""
    working_dir: Path = Path(project_dir, "demultiplexed-runs")
    working_dir.mkdir(parents=True)
    return working_dir


@pytest.fixture(name="demultiplexed_flow_cell_working_directory")
def fixture_demultiplexed_flow_cell_working_directory(
    demux_results_not_finished_dir: Path,
    demultiplexed_flow_cells_working_directory: Path,
    flow_cell_full_name: str,
) -> Path:
    """Copy the content of a demultiplexed but not finished directory to a temporary location."""
    source: Path = Path(demux_results_not_finished_dir, flow_cell_full_name)
    destination: Path = Path(demultiplexed_flow_cells_working_directory, flow_cell_full_name)
    shutil.copytree(src=source, dst=destination)
    return destination


@pytest.fixture(name="demultiplexed_flow_cell_finished_working_directory")
def fixture_demultiplexed_flow_cell_finished_working_directory(
    demultiplexed_runs: Path,
    demultiplexed_flow_cells_working_directory: Path,
    flow_cell_full_name: str,
) -> Path:
    """Copy the content of a demultiplexed but not finished directory to a temporary location."""
    source: Path = Path(demultiplexed_runs, flow_cell_full_name)
    destination: Path = Path(demultiplexed_flow_cells_working_directory, flow_cell_full_name)
    shutil.copytree(src=source, dst=destination)
    return destination


@pytest.fixture(name="flow_cell_working_directory")
def fixture_flow_cell_working_directory(
    novaseq_dir: Path, flow_cell_runs_working_directory: Path
) -> Path:
    """Return the path to a working directory that will be deleted after test is run.

    This is a path to a flow cell directory with the run parameters present.
    """
    working_dir: Path = Path(flow_cell_runs_working_directory, novaseq_dir.name)
    working_dir.mkdir(parents=True)
    existing_flow_cell: FlowCell = FlowCell(flow_cell_path=novaseq_dir)
    working_flow_cell: FlowCell = FlowCell(flow_cell_path=working_dir)
    shutil.copy(
        existing_flow_cell.run_parameters_path.as_posix(),
        working_flow_cell.run_parameters_path.as_posix(),
    )
    return working_dir


@pytest.fixture(name="flow_cell_working_directory_bcl2fastq")
def fixture_flow_cell_working_directory_bcl2fastq(
    flow_cell_dir_bcl2fastq: Path, flow_cell_runs_working_directory_bcl2fastq: Path
) -> Path:
    """Return the path to a working directory that will be deleted after test is run.

    This is a path to a flow cell directory with the run parameters present.
    """
    working_dir: Path = Path(
        flow_cell_runs_working_directory_bcl2fastq, flow_cell_dir_bcl2fastq.name
    )
    working_dir.mkdir(parents=True)
    existing_flow_cell: FlowCell = FlowCell(flow_cell_path=flow_cell_dir_bcl2fastq)
    working_flow_cell: FlowCell = FlowCell(flow_cell_path=working_dir)
    shutil.copy(
        existing_flow_cell.run_parameters_path.as_posix(),
        working_flow_cell.run_parameters_path.as_posix(),
    )
    return working_dir


@pytest.fixture(name="flow_cell_working_directory_dragen")
def fixture_flow_cell_working_directory_dragen(
    flow_cell_dir_dragen: Path, flow_cell_runs_working_directory_dragen: Path
) -> Path:
    """Return the path to a working directory that will be deleted after test is run.

    This is a path to a flow cell directory with the run parameters present.
    """
    working_dir: Path = Path(flow_cell_runs_working_directory_dragen, flow_cell_dir_dragen.name)
    working_dir.mkdir(parents=True)
    existing_flow_cell: FlowCell = FlowCell(flow_cell_path=flow_cell_dir_dragen)
    working_flow_cell: FlowCell = FlowCell(flow_cell_path=working_dir)
    shutil.copy(
        existing_flow_cell.run_parameters_path.as_posix(),
        working_flow_cell.run_parameters_path.as_posix(),
    )
    return working_dir


@pytest.fixture(name="flow_cell_working_directory_no_run_parameters")
def fixture_flow_cell_working_directory_no_run_parameters(
    novaseq_dir: Path, flow_cell_runs_working_directory: Path
) -> Path:
    """This is a path to a flow cell directory with the run parameters missing."""
    working_dir: Path = Path(flow_cell_runs_working_directory, novaseq_dir.name)
    working_dir.mkdir(parents=True)
    return working_dir


@pytest.fixture(name="demultiplex_ready_flow_cell")
def fixture_demultiplex_ready_flow_cell(
    flow_cell_working_directory: Path, novaseq_dir: Path
) -> Path:
    """Return the path to a working directory that is ready for demultiplexing.

    This is a path to a flow cell directory with all the files necessary to start demultiplexing present.
    """
    existing_flow_cell: FlowCell = FlowCell(flow_cell_path=novaseq_dir)
    working_flow_cell: FlowCell = FlowCell(flow_cell_path=flow_cell_working_directory)
    shutil.copy(
        existing_flow_cell.sample_sheet_path.as_posix(),
        working_flow_cell.sample_sheet_path.as_posix(),
    )
    shutil.copy(
        str(DemultiplexingAPI.get_stderr_logfile(existing_flow_cell)),
        str(DemultiplexingAPI.get_stderr_logfile(working_flow_cell)),
    )
    working_flow_cell.copy_complete_path.touch()
    working_flow_cell.rta_complete_path.touch()
    return flow_cell_working_directory


@pytest.fixture(name="demultiplex_ready_flow_cell_bcl2fastq")
def fixture_demultiplex_ready_flow_cell_bcl2fastq(
    flow_cell_working_directory_bcl2fastq: Path, flow_cell_dir_bcl2fastq: Path
) -> Path:
    """Return the path to a working directory that is ready for demultiplexing.

    This is a path to a flow cell directory with all the files necessary to start demultiplexing present.
    """
    existing_flow_cell: FlowCell = FlowCell(flow_cell_path=flow_cell_dir_bcl2fastq)
    working_flow_cell: FlowCell = FlowCell(flow_cell_path=flow_cell_working_directory_bcl2fastq)
    shutil.copy(
        existing_flow_cell.sample_sheet_path.as_posix(),
        working_flow_cell.sample_sheet_path.as_posix(),
    )
    shutil.copy(
        str(DemultiplexingAPI.get_stderr_logfile(existing_flow_cell)),
        str(DemultiplexingAPI.get_stderr_logfile(working_flow_cell)),
    )
    working_flow_cell.copy_complete_path.touch()
    working_flow_cell.rta_complete_path.touch()
    return flow_cell_working_directory_bcl2fastq


@pytest.fixture(name="demultiplex_ready_flow_cell_dragen")
def fixture_demultiplex_ready_flow_cell_dragen(
    flow_cell_working_directory_dragen: Path, flow_cell_dir_dragen: Path
) -> Path:
    """Return the path to a working directory that is ready for demultiplexing.

    This is a path to a flow cell directory with all the files necessary to start demultiplexing present.
    """
    existing_flow_cell: FlowCell = FlowCell(
        flow_cell_path=flow_cell_dir_dragen, bcl_converter="dragen"
    )
    working_flow_cell: FlowCell = FlowCell(
        flow_cell_path=flow_cell_working_directory_dragen, bcl_converter="dragen"
    )
    shutil.copy(
        existing_flow_cell.sample_sheet_path.as_posix(),
        working_flow_cell.sample_sheet_path.as_posix(),
    )
    shutil.copy(
        str(DemultiplexingAPI.get_stderr_logfile(existing_flow_cell)),
        str(DemultiplexingAPI.get_stderr_logfile(working_flow_cell)),
    )
    working_flow_cell.copy_complete_path.touch()
    working_flow_cell.rta_complete_path.touch()
    return flow_cell_working_directory_dragen


@pytest.fixture(name="sample_sheet_context")
def fixture_sample_sheet_context(cg_context: CGConfig, lims_api: LimsAPI) -> CGConfig:
    """Return cg context with an added lims API."""
    cg_context.lims_api_ = lims_api
    return cg_context


@pytest.fixture(name="demultiplex_configs")
def fixture_demultiplex_configs(
    flow_cell_runs_working_directory: Path,
    demultiplexed_flow_cells_working_directory: Path,
    demultiplex_fixtures: Path,
) -> dict:
    """Return demultiplex configs."""
    demultiplexed_flow_cells_working_directory.mkdir(parents=True, exist_ok=True)
    return {
        "demultiplex": {
            "out_dir": demultiplexed_flow_cells_working_directory.as_posix(),
            "run_dir": flow_cell_runs_working_directory.as_posix(),
            "slurm": {"account": "test", "mail_user": "testuser@github.se"},
        }
    }


@pytest.fixture(name="demultiplexing_api")
def fixture_demultiplexing_api(
    demultiplex_configs: dict, sbatch_process: Process
) -> DemultiplexingAPI:
    """Return demultiplex API."""
    demux_api = DemultiplexingAPI(config=demultiplex_configs)
    demux_api.slurm_api.process = sbatch_process
    return demux_api


@pytest.fixture(name="demultiplex_context")
def fixture_demultiplex_context(
    demultiplexing_api: DemultiplexingAPI,
    stats_api: StatsAPI,
    real_housekeeper_api: HousekeeperAPI,
    cg_context: CGConfig,
) -> CGConfig:
    """Return cg context witha demultiplex context."""
    cg_context.demultiplex_api_ = demultiplexing_api
    cg_context.cg_stats_api_ = stats_api
    cg_context.housekeeper_api_ = real_housekeeper_api
    return cg_context



from pathlib import Path

import pytest

from tests.cli.demultiplex.conftest import fixture_demultiplexed_flow_cells_working_directory
from tests.models.demultiplexing.conftest import (
    fixture_flowcell_path,
    fixture_flow_cell_runs,
)


@pytest.fixture(name="correct_flow_cell")
def fixture_correct_flow_cell() -> str:
    """Correct flow cell name"""
    return "201203_A00689_0200_AHVKJCDRXX"


@pytest.fixture(name="incorrect_flow_cell_too_long")
def fixture_incorrect_flow_cell_too_long() -> str:
    """Incorrect flow cell name"""
    return "201203_A00689_0200_AHVKJCDRXX_r"


@pytest.fixture(name="incorrect_flow_cell_extension")
def fixture_incorrect_flow_cell_extension() -> str:
    """Incorrect flow cell name"""
    return "201203_A00689_0200_AHVKJCDRXX.someextension"


@pytest.fixture(name="incorrect_flow_cell_name")
def fixture_incorrect_flow_cell_name() -> str:
    """Incorrect flow cell name"""
    return "201203_A00689_0200_AZZZZZZZZZ"


@pytest.fixture(name="correct_flow_cell_path")
def fixture_correct_flow_cell_path(
    demultiplexed_flow_cells_working_directory, correct_flow_cell
) -> Path:
    """Full path to a correctly named flow cell directory in demultiplexed-runs"""
    return Path(demultiplexed_flow_cells_working_directory, correct_flow_cell)


@pytest.fixture(name="incorrect_flow_cell_path_too_long")
def fixture_incorrect_flow_cell_path_too_long(
    demultiplexed_flow_cells_working_directory, incorrect_flow_cell_too_long
) -> Path:
    """Full path to an incorrectly named flow cell directory in demultiplexed-runs"""
    return Path(demultiplexed_flow_cells_working_directory, incorrect_flow_cell_too_long)


@pytest.fixture(name="incorrect_flow_cell_path_extension")
def fixture_incorrect_flow_cell_path_extension(
    demultiplexed_flow_cells_working_directory, incorrect_flow_cell_extension
) -> Path:
    """Full path to an incorrectly named flow cell directory in demultiplexed-runs"""
    return Path(demultiplexed_flow_cells_working_directory, incorrect_flow_cell_extension)


@pytest.fixture(name="nonexistent_flow_cell_path")
def fixture_nonexistent_flow_cell_path(
    demultiplexed_flow_cells_working_directory, incorrect_flow_cell_name
) -> Path:
    """Full path to an incorrectly named flow cell directory in demultiplexed-runs"""
    return Path(demultiplexed_flow_cells_working_directory, incorrect_flow_cell_name)



"""Fixtures for cli workflow taxprofiler tests"""

import pytest
from cg.meta.workflow.taxprofiler import TaxprofilerAnalysisAPI
from cg.models.cg_config import CGConfig


@pytest.fixture(scope="function", name="taxprofiler_context")
def fixture_taxprofiler_context(
    cg_context: CGConfig,
) -> CGConfig:
    """Context to use in cli."""
    cg_context.meta_apis["analysis_api"] = TaxprofilerAnalysisAPI(config=cg_context)
    return cg_context



"""Tests for cleaning flow cell run directories using
cg.meta.clean.flow_cell_run_directories.RunDirFlowCell."""
from unittest import mock

from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.housekeeper_tags import SequencingFileTag
from cg.meta.clean.flow_cell_run_directories import RunDirFlowCell
from tests.mocks.hk_mock import MockBundle


@mock.patch("cg.apps.housekeeper.hk.HousekeeperAPI")
@mock.patch("cg.store.Store")
def test_age(
    mock_statusdb,
    mock_hk,
    flow_cell_path,
    timestamp_yesterday,
):
    # GIVEN a flow cell with a sequenced date:
    flow_cell: RunDirFlowCell = RunDirFlowCell(flow_cell_path, mock_statusdb, mock_hk)
    flow_cell._sequenced_date = timestamp_yesterday

    # WHEN determining the age of a flow cell
    result = flow_cell.age

    # THEN the age property of the flow cell should be set to 1 day
    assert result == 1


@mock.patch("cg.apps.housekeeper.hk.HousekeeperAPI")
@mock.patch("cg.store.Store")
def test_sequenced_date_from_statusdb(
    mock_statusdb,
    mock_hk,
    flow_cell_path,
    timestamp_yesterday,
):
    # GIVEN a flow cell with a sequenced_at date in statusdb
    flow_cell: RunDirFlowCell = RunDirFlowCell(flow_cell_path, mock_statusdb, mock_hk)
    mock_statusdb.get_flow_cell_by_name.return_value.sequenced_at = timestamp_yesterday

    # WHEN determining the age of a flow cell
    result = flow_cell.sequenced_date

    # THEN the sequenced_date property of the flow cell should be set to the sequenced_at date in
    # statusdb
    assert result == timestamp_yesterday


@mock.patch("cg.apps.housekeeper.hk.HousekeeperAPI")
@mock.patch("cg.store.Store")
def test_sequenced_date_from_run_name(
    mock_statusdb,
    mock_hk,
    flow_cell_path,
):
    # GIVEN a flow cell that does not exist in statusdb
    flow_cell: RunDirFlowCell = RunDirFlowCell(flow_cell_path, mock_statusdb, mock_hk)
    mock_statusdb.get_flow_cell_by_name.return_value = None

    # WHEN determining the age of a flow cell
    result = flow_cell.sequenced_date

    # THEN the sequenced_date property of the flow cell should be derived from the run name of
    # the flow cell
    assert result == flow_cell.derived_date


@mock.patch("cg.apps.housekeeper.hk.HousekeeperAPI")
@mock.patch("cg.store.Store")
def test_archive_sample_sheet_no_bundle(mock_statusdb, mock_hk, flow_cell_path, novaseq_dir):
    # GIVEN a flow cell
    flow_cell: RunDirFlowCell = RunDirFlowCell(flow_cell_path, mock_statusdb, mock_hk)
    # GIVEN a sample sheet connected to the flow cell
    flow_cell.sample_sheet_path = novaseq_dir / DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME
    # GIVEN there is no bundle for the flow cell
    mock_hk.bundle.return_value = None
    # GIVEN the sample sheet does not exist in Housekeeper
    mock_hk.files.return_value.first.return_value = None
    mock_hk.get_files.return_value.first.return_value.is_included = False

    # WHEN archiving the sample sheet
    flow_cell.archive_sample_sheet()

    # THEN a bundle and version should be created for the flow cell
    flow_cell.hk.create_new_bundle_and_version.assert_called_once_with(name=flow_cell.id)

    # THEN the sample sheet should be be added to Housekeeper
    flow_cell.hk.add_and_include_file_to_latest_version.assert_called_once_with(
        bundle_name=flow_cell.id,
        file=flow_cell.sample_sheet_path,
        tags=[SequencingFileTag.ARCHIVED_SAMPLE_SHEET, flow_cell.id],
    )


@mock.patch("cg.apps.housekeeper.hk.HousekeeperAPI")
@mock.patch("cg.store.Store")
def test_archive_sample_sheet_included(mock_statusdb, mock_hk, flow_cell_path, novaseq_dir, caplog):
    """Test archive of a sample sheet when it has been already included in HK."""

    # GIVEN a flow cell
    flow_cell: RunDirFlowCell = RunDirFlowCell(flow_cell_path, mock_statusdb, mock_hk)

    # GIVEN a sample sheet connected to the flow cell
    flow_cell.sample_sheet_path = novaseq_dir / DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME

    # GIVEN the sample sheet does exist in Housekeeper
    mock_hk.get_file_from_latest_version.return_value = True

    # WHEN archiving the sample sheet
    flow_cell.archive_sample_sheet()

    # THEN the sample sheet should not be included again
    assert "Sample sheet already included!" in caplog.text



"""Tests for the utils module"""

from cg.utils.utils import get_string_from_list_by_pattern


def test_get_string_from_list_by_pattern():
    """Tests element extraction from a list given a string pattern."""

    # GIVEN a specific pattern and a list containing it
    pattern = "pattern"
    list_of_string = ["I do contain the pattern", "test string", "another random string"]

    # WHEN calling the filtering method
    extracted_string: str = get_string_from_list_by_pattern(list_of_string, pattern)

    # THEN the extracted string should match the one containing the pattern
    assert extracted_string == list_of_string[0]


def test_get_string_from_list_by_pattern_not_found():
    """Tests element extraction from a list given a string pattern that does not exist in the list."""

    # GIVEN a specific pattern and a list not containing it
    pattern = "I am not in the list"
    list_of_string = ["This", "is", "a", "random", "list"]

    # WHEN calling the filtering method
    extracted_string: str = get_string_from_list_by_pattern(list_of_string, pattern)

    # THEN None should be returned
    assert extracted_string is None



"""PDC related constants"""
from cg.utils.enums import ListEnum, IntEnum


class DSMCParameters(ListEnum):
    ARCHIVE_COMMAND: list = ["archive"]
    QUERY_COMMAND: list = ["q", "archive"]
    RETRIEVE_COMMAND: list = ["retrieve", "-replace=yes"]


class PDCExitCodes(IntEnum):
    """Exit codes for PDC commands"""

    SUCCESS: int = 0
    NO_FILES_FOUND: int = 8



"""Module for reading and writing comma separated values (CSV) formatted files."""
import csv
import io
from typing import Any, List, Tuple
from pathlib import Path


def read_csv(file_path: Path) -> List[List[str]]:
    """Read content in a CSV file."""
    with open(file_path, "r") as file:
        csv_reader = csv.reader(file)
        return [row for row in csv_reader]


def read_csv_stream(stream: str) -> List[List[str]]:
    """Read CSV formatted stream."""
    csv_reader = csv.reader(stream.splitlines())
    return [row for row in csv_reader]


def write_csv(content: List[List[Any]], file_path: Path) -> None:
    """Write content to a CSV file."""
    with open(file_path, "w", newline="") as file:
        csv_writer = csv.writer(file)
        for row in content:
            csv_writer.writerow(row)


def write_csv_stream(content: List[List[Any]]) -> str:
    """Write content to a CSV stream."""
    csv_stream = io.StringIO()
    csv_writer = csv.writer(csv_stream, lineterminator="\n")
    for row in content:
        csv_writer.writerow(row)
    return csv_stream.getvalue()



import logging
from pathlib import Path
from typing import Dict, List, Union
from pydantic import parse_obj_as
from typing_extensions import Literal

from cg.apps.demultiplex.sample_sheet.models import (
    NovaSeqSample,
    SampleSheet,
    SampleBcl2Fastq,
    SampleDragen,
)
from cg.constants.demultiplexing import BclConverter
from cg.exc import SampleSheetError

LOG = logging.getLogger(__name__)


def validate_samples_are_unique(samples: List[NovaSeqSample]) -> None:
    """Validate that each sample only exists once."""
    sample_ids: set = set()
    for sample in samples:
        sample_id: str = sample.sample_id.split("_")[0]
        if sample_id in sample_ids:
            message: str = f"Sample {sample.sample_id} exists multiple times in sample sheet"
            LOG.warning(message)
            raise SampleSheetError(message)
        sample_ids.add(sample_id)


def get_samples_by_lane(samples: List[NovaSeqSample]) -> Dict[int, List[NovaSeqSample]]:
    """Group and return samples by lane."""
    LOG.info("Order samples by lane")
    sample_by_lane: Dict[int, List[NovaSeqSample]] = {}
    for sample in samples:
        if sample.lane not in sample_by_lane:
            sample_by_lane[sample.lane] = []
        sample_by_lane[sample.lane].append(sample)
    return sample_by_lane


def validate_samples_unique_per_lane(samples: List[NovaSeqSample]) -> None:
    """Validate that each sample only exists once per lane in a sample sheet."""

    sample_by_lane: Dict[int, List[NovaSeqSample]] = get_samples_by_lane(samples)
    for lane, lane_samples in sample_by_lane.items():
        LOG.info(f"Validate that samples are unique in lane {lane}")
        validate_samples_are_unique(samples=lane_samples)


def get_raw_samples(sample_sheet: str) -> List[Dict[str, str]]:
    """Return the samples in a sample sheet as a list of dictionaries."""
    sample_sheet_rows: List[str] = sample_sheet.split("\n")
    header: List[str] = []
    raw_samples: List[Dict[str, str]] = []
    for line in sample_sheet_rows:
        # Skip lines that are too short to contain samples
        if not len(line) > 5:
            continue
        line = line.strip()
        # Check if we are on the header row
        if line.startswith("FCID"):
            header = line.split(",")
            continue
        if not header:
            continue
        raw_samples.append(dict(zip(header, line.split(","))))
    if not header:
        message = "Could not find header in sample sheet"
        LOG.warning(message)
        raise SampleSheetError(message)
    if not raw_samples:
        message = "Could not find any samples in sample sheet"
        LOG.warning(message)
        raise SampleSheetError(message)
    return raw_samples


def get_sample_sheet(
    sample_sheet: str, sheet_type: Literal["2500", "SP", "S2", "S4"], bcl_converter: str
) -> SampleSheet:
    """Parse and validate a sample sheet.

    return the information as a SampleSheet object
    """
    novaseq_sample = {BclConverter.BCL2FASTQ: SampleBcl2Fastq, BclConverter.DRAGEN: SampleDragen}
    raw_samples: List[Dict[str, str]] = get_raw_samples(sample_sheet)
    sample_type: Union[SampleBcl2Fastq, SampleDragen] = novaseq_sample[bcl_converter]
    samples = parse_obj_as(List[sample_type], raw_samples)
    validate_samples_unique_per_lane(samples)
    return SampleSheet(type=sheet_type, samples=samples)


def get_sample_sheet_from_file(
    infile: Path, sheet_type: Literal["2500", "SP", "S2", "S4"], bcl_converter: str
) -> SampleSheet:
    """Parse and validate a sample sheet from file."""
    with open(infile, "r") as csv_file:
        sample_sheet: SampleSheet = get_sample_sheet(
            sample_sheet=csv_file.read(), sheet_type=sheet_type, bcl_converter=bcl_converter
        )

    return sample_sheet



"""Module for Rnafusion Analysis API."""

import logging
from pathlib import Path
from typing import Dict, List, Optional

from pydantic import ValidationError

from cg import resources
from cg.constants import Pipeline
from cg.constants.constants import FileFormat, WorkflowManager
from cg.constants.nextflow import NFX_READ1_HEADER, NFX_READ2_HEADER, NFX_SAMPLE_HEADER
from cg.constants.rnafusion import (
    RNAFUSION_METRIC_CONDITIONS,
    RNAFUSION_SAMPLESHEET_HEADERS,
    RNAFUSION_STRANDEDNESS_HEADER,
    RnafusionDefaults,
)
from cg.io.controller import WriteFile
from cg.io.json import read_json
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.fastq import RnafusionFastqHandler
from cg.meta.workflow.nextflow_common import NextflowAnalysisAPI
from cg.meta.workflow.tower_common import TowerAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.deliverables.metric_deliverables import MetricsBase, MultiqcDataJson
from cg.models.nextflow.deliverables import NextflowDeliverables, replace_dict_values
from cg.models.rnafusion.rnafusion_sample import RnafusionSample
from cg.utils import Process

LOG = logging.getLogger(__name__)


class RnafusionAnalysisAPI(AnalysisAPI):
    """Handles communication between RNAFUSION processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        pipeline: Pipeline = Pipeline.RNAFUSION,
    ):
        super().__init__(config=config, pipeline=pipeline)
        self.root_dir: str = config.rnafusion.root
        self.nfcore_pipeline_path: str = config.rnafusion.pipeline_path
        self.references: str = config.rnafusion.references
        self.profile: str = config.rnafusion.profile
        self.conda_env: str = config.rnafusion.conda_env
        self.conda_binary: str = config.rnafusion.conda_binary
        self.tower_binary_path: str = config.rnafusion.tower_binary_path
        self.tower_pipeline: str = config.rnafusion.tower_pipeline
        self.account: str = config.rnafusion.slurm.account
        self.email: str = config.rnafusion.slurm.mail_user
        self.compute_env: str = config.rnafusion.compute_env
        self.revision: str = config.rnafusion.revision

    @property
    def root(self) -> str:
        return self.root_dir

    @property
    def fastq_handler(self):
        return RnafusionFastqHandler

    @property
    def process(self):
        if not self._process:
            self._process = Process(
                binary=self.tower_binary_path,
            )
        return self._process

    @process.setter
    def process(self, process: Process):
        self._process = process

    def get_profile(self, profile: Optional[str] = None) -> str:
        if profile:
            return profile
        return self.profile

    def get_workflow_manager(self) -> str:
        """Get workflow manager for rnafusion."""
        return WorkflowManager.Tower.value

    def get_case_config_path(self, case_id):
        return NextflowAnalysisAPI.get_case_config_path(case_id=case_id, root_dir=self.root_dir)

    def verify_analysis_finished(self, case_id):
        return NextflowAnalysisAPI.verify_analysis_finished(case_id=case_id, root_dir=self.root_dir)

    @staticmethod
    def build_samplesheet_content(
        case_id: str, fastq_r1: List[str], fastq_r2: List[str], strandedness: str
    ) -> Dict[str, List[str]]:
        """Build samplesheet headers and lists"""
        try:
            RnafusionSample(
                sample=case_id, fastq_r1=fastq_r1, fastq_r2=fastq_r2, strandedness=strandedness
            )
        except ValidationError as error:
            LOG.error(error)
            raise ValueError

        samples_full_list: list = []
        strandedness_full_list: list = []
        # Complete sample and strandedness lists to the same length as fastq_r1:
        for _ in range(len(fastq_r1)):
            samples_full_list.append(case_id)
            strandedness_full_list.append(strandedness)

        samplesheet_content: dict = {
            NFX_SAMPLE_HEADER: samples_full_list,
            NFX_READ1_HEADER: fastq_r1,
            NFX_READ2_HEADER: fastq_r2,
            RNAFUSION_STRANDEDNESS_HEADER: strandedness_full_list,
        }
        return samplesheet_content

    def write_samplesheet(self, case_id: str, strandedness: str, dry_run: bool = False) -> None:
        """Write sample sheet for rnafusion analysis in case folder."""
        case_obj = self.status_db.get_case_by_internal_id(internal_id=case_id)
        if len(case_obj.links) != 1:
            raise NotImplementedError(
                "Case objects are assumed to be related to a single sample (one link)"
            )

        for link in case_obj.links:
            sample_metadata: List[str] = self.gather_file_metadata_for_sample(link.sample)
            fastq_r1: List[str] = NextflowAnalysisAPI.extract_read_files(1, sample_metadata)
            fastq_r2: List[str] = NextflowAnalysisAPI.extract_read_files(2, sample_metadata)
            samplesheet_content: Dict[str, List[str]] = self.build_samplesheet_content(
                case_id, fastq_r1, fastq_r2, strandedness
            )
            LOG.info(samplesheet_content)
            if dry_run:
                continue
            NextflowAnalysisAPI.create_samplesheet_csv(
                samplesheet_content=samplesheet_content,
                headers=RNAFUSION_SAMPLESHEET_HEADERS,
                config_path=NextflowAnalysisAPI.get_case_config_path(
                    case_id=case_id, root_dir=self.root_dir
                ),
            )

    def write_params_file(
        self, case_id: str, genomes_base: Optional[Path] = None, dry_run: bool = False
    ) -> None:
        """Write params-file for rnafusion analysis in case folder."""
        default_options: Dict[str, str] = self.get_default_parameters(case_id=case_id)
        if genomes_base:
            default_options["genomes_base"] = genomes_base
        LOG.info(default_options)
        if dry_run:
            return
        NextflowAnalysisAPI.write_nextflow_yaml(
            content=default_options,
            file_path=NextflowAnalysisAPI.get_params_file_path(
                case_id=case_id, root_dir=self.root_dir
            ),
        )

    def get_trailblazer_config_path(self, case_id: str) -> Path:
        """Return the path to a trailblazer config file containing Tower IDs."""
        return Path(self.root_dir, case_id, "tower_ids.yaml")

    def write_trailblazer_config(self, case_id: str, tower_id: str) -> None:
        """Write Tower IDs to a .YAML file used as the trailblazer config."""
        config_path = self.get_trailblazer_config_path(case_id=case_id)
        LOG.info(f"Writing Tower ID to {config_path.as_posix()}")
        WriteFile.write_file_from_content(
            content={case_id: [tower_id]},
            file_format=FileFormat.YAML,
            file_path=config_path,
        )

    def get_references_path(self, genomes_base: Optional[Path] = None) -> Path:
        if genomes_base:
            return genomes_base.absolute()
        return Path(self.references).absolute()

    def get_default_parameters(self, case_id: str) -> Dict:
        """Returns a dictionary with default RNAFusion parameters."""
        return {
            "input": NextflowAnalysisAPI.get_input_path(
                case_id=case_id, root_dir=self.root_dir
            ).as_posix(),
            "outdir": NextflowAnalysisAPI.get_outdir_path(
                case_id=case_id, root_dir=self.root_dir
            ).as_posix(),
            "genomes_base": self.get_references_path().as_posix(),
            "trim": RnafusionDefaults.TRIM,
            "fastp_trim": RnafusionDefaults.FASTP_TRIM,
            "trim_tail": RnafusionDefaults.TRIM_TAIL,
            "fusioninspector_filter": RnafusionDefaults.FUSIONINSPECTOR_FILTER,
            "fusionreport_filter": RnafusionDefaults.FUSIONREPORT_FILTER,
            "all": RnafusionDefaults.ALL,
            "pizzly": RnafusionDefaults.PIZZLY,
            "squid": RnafusionDefaults.SQUID,
            "starfusion": RnafusionDefaults.STARFUSION,
            "fusioncatcher": RnafusionDefaults.FUSIONCATCHER,
            "arriba": RnafusionDefaults.ARRIBA,
            "cram": RnafusionDefaults.CRAM,
            "priority": self.account,
            "clusterOptions": f"--qos={self.get_slurm_qos_for_case(case_id=case_id)}",
        }

    def config_case(
        self,
        case_id: str,
        strandedness: str,
        genomes_base: Path,
        dry_run: bool,
    ) -> None:
        """Create sample sheet file for RNAFUSION analysis."""
        NextflowAnalysisAPI.make_case_folder(
            case_id=case_id, root_dir=self.root_dir, dry_run=dry_run
        )
        LOG.info("Generating samplesheet")
        self.write_samplesheet(case_id=case_id, strandedness=strandedness, dry_run=dry_run)
        LOG.info("Generating parameters file")
        self.write_params_file(case_id=case_id, genomes_base=genomes_base, dry_run=dry_run)
        if dry_run:
            LOG.info("Dry run: Config files will not be written")
            return

        LOG.info("Configs files written")

    def run_analysis(
        self, case_id: str, command_args: dict, use_nextflow: bool, dry_run: bool = False
    ) -> None:
        """Execute RNAFUSION run analysis with given options."""
        if use_nextflow:
            self.process = Process(
                binary=self.config.rnafusion.binary_path,
                environment=self.conda_env,
                conda_binary=self.conda_binary,
                launch_directory=NextflowAnalysisAPI.get_case_path(
                    case_id=case_id, root_dir=self.root_dir
                ),
            )
            LOG.info("Pipeline will be executed using nextflow")
            parameters: List[str] = NextflowAnalysisAPI.get_nextflow_run_parameters(
                case_id=case_id,
                pipeline_path=self.nfcore_pipeline_path,
                root_dir=self.root_dir,
                command_args=command_args,
            )
            self.process.export_variables(
                export=NextflowAnalysisAPI.get_variables_to_export(
                    case_id=case_id, root_dir=self.root_dir
                ),
            )

            command = self.process.get_command(parameters=parameters)
            LOG.info(f"{command}")
            sbatch_number: int = NextflowAnalysisAPI.execute_head_job(
                case_id=case_id,
                root_dir=self.root_dir,
                slurm_account=self.account,
                email=self.email,
                qos=self.get_slurm_qos_for_case(case_id=case_id),
                commands=command,
                dry_run=dry_run,
            )
            LOG.info(f"Nextflow head job running as job {sbatch_number}")

        else:
            LOG.info("Pipeline will be executed using tower")
            parameters: List[str] = TowerAnalysisAPI.get_tower_launch_parameters(
                tower_pipeline=self.tower_pipeline,
                command_args=command_args,
            )
            self.process.run_command(parameters=parameters, dry_run=dry_run)
            if self.process.stderr:
                LOG.error(self.process.stderr)
            if not dry_run:
                tower_id = TowerAnalysisAPI.get_tower_id(stdout_lines=self.process.stdout_lines())
                self.write_trailblazer_config(case_id=case_id, tower_id=tower_id)
            LOG.info(self.process.stdout)

    def verify_case_config_file_exists(self, case_id: str, dry_run: bool = False) -> None:
        NextflowAnalysisAPI.verify_case_config_file_exists(
            case_id=case_id, root_dir=self.root_dir, dry_run=dry_run
        )

    def get_deliverables_file_path(self, case_id: str) -> Path:
        return NextflowAnalysisAPI.get_deliverables_file_path(
            case_id=case_id, root_dir=self.root_dir
        )

    def get_pipeline_version(self, case_id: str) -> str:
        return NextflowAnalysisAPI.get_pipeline_version(
            case_id=case_id, root_dir=self.root_dir, pipeline=self.pipeline
        )

    def verify_deliverables_file_exists(self, case_id: str) -> None:
        NextflowAnalysisAPI.verify_deliverables_file_exists(case_id=case_id, root_dir=self.root_dir)

    def report_deliver(self, case_id: str) -> None:
        """Get a deliverables file template from resources, parse it and, then write the deliverables file."""
        deliverables_content: dict = NextflowAnalysisAPI.get_template_deliverables_file_content(
            resources.RNAFUSION_BUNDLE_FILENAMES_PATH
        )
        try:
            for index, deliver_file in enumerate(deliverables_content):
                NextflowDeliverables(deliverables=deliver_file)
                deliverables_content[index] = replace_dict_values(
                    NextflowAnalysisAPI.get_replace_map(case_id=case_id, root_dir=self.root_dir),
                    deliver_file,
                )
        except ValidationError as error:
            LOG.error(error)
            raise ValueError
        NextflowAnalysisAPI.make_case_folder(case_id=case_id, root_dir=self.root_dir)
        NextflowAnalysisAPI.write_deliverables_bundle(
            deliverables_content=NextflowAnalysisAPI.add_bundle_header(
                deliverables_content=deliverables_content
            ),
            file_path=NextflowAnalysisAPI.get_deliverables_file_path(
                case_id=case_id, root_dir=self.root_dir
            ),
        )
        LOG.info(
            "Writing deliverables file in "
            + str(
                NextflowAnalysisAPI.get_deliverables_file_path(
                    case_id=case_id, root_dir=self.root_dir
                )
            )
        )

    def get_multiqc_json_path(self, case_id: str) -> Path:
        """Return the path of the multiqc_data.json file."""
        return Path(self.root_dir, case_id, "multiqc", "multiqc_data", "multiqc_data.json")

    def get_metrics_deliverables_path(self, case_id: str) -> Path:
        """Return a path where the <case>_metrics_deliverables.yaml file should be located."""
        return Path(self.root_dir, case_id, f"{case_id}_metrics_deliverables.yaml")

    def get_multiqc_json_metrics(self, case_id: str) -> List[MetricsBase]:
        """Get a multiqc_data.json file and returns metrics and values formatted."""
        multiqc_json: MultiqcDataJson = MultiqcDataJson(
            **read_json(file_path=self.get_multiqc_json_path(case_id=case_id))
        )
        metrics_values: Dict = {}
        for key in multiqc_json.report_general_stats_data:
            if case_id in key:
                metrics_values.update(list(key.values())[0])
        return [
            MetricsBase(
                header=None,
                id=case_id,
                input="multiqc_data.json",
                name=metric_name,
                step="multiqc",
                value=metric_value,
                condition=RNAFUSION_METRIC_CONDITIONS.get(metric_name, None),
            )
            for metric_name, metric_value in metrics_values.items()
        ]

    def write_metrics_deliverables(self, case_id: str, dry_run: bool = False):
        """Write <case>_metrics_deliverables.yaml file."""
        metrics_deliverables_path: Path = self.get_metrics_deliverables_path(case_id=case_id)
        if dry_run:
            LOG.info(
                f"Dry run: metrics deliverables file would be written to {metrics_deliverables_path.as_posix()}"
            )
            return
        LOG.info(f"Writing metrics deliverables file to {metrics_deliverables_path.as_posix()}")
        WriteFile.write_file_from_content(
            content={
                "metrics": [
                    metric.dict() for metric in self.get_multiqc_json_metrics(case_id=case_id)
                ]
            },
            file_format=FileFormat.YAML,
            file_path=metrics_deliverables_path,
        )



"""This script tests the cli methods to add families to status-db"""

from cg.cli.status import status_of_cases, status
from cg.store import Store
from click.testing import CliRunner
from cg.models.cg_config import CGConfig
from tests.store_helpers import StoreHelpers


def test_lists_sample_in_unreceived_samples(
    cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers
):
    """Test to that cases displays case in database"""

    # GIVEN a database with a case
    base_store: Store = base_context.status_db
    case = helpers.add_case(base_store)
    sample1 = helpers.add_sample(base_store, "sample1")
    helpers.add_relationship(base_store, case=case, sample=sample1)

    # WHEN listing cases
    result = cli_runner.invoke(status_of_cases, ["-o", "count"], obj=base_context)
    result2 = cli_runner.invoke(status, ["cases", "-o", "count"], obj=base_context)

    # THEN the case should be listed
    assert result.exit_code == 0
    assert case.internal_id in result.output
    assert "0/1" in result.output
    assert result.output == result2.output


def test_lists_samples_in_unreceived_samples(
    cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers
):
    """Test to that cases displays case in database"""

    # GIVEN a database with a case
    base_store: Store = base_context.status_db
    case = helpers.add_case(base_store)
    sample1 = helpers.add_sample(base_store, "sample1")
    sample2 = helpers.add_sample(base_store, "sample2")
    helpers.add_relationship(base_store, case=case, sample=sample1)
    helpers.add_relationship(base_store, case=case, sample=sample2)

    # WHEN listing cases
    result = cli_runner.invoke(status, ["cases", "-o", "count"], obj=base_context)

    # THEN the case should be listed
    assert result.exit_code == 0
    assert case.internal_id in result.output
    assert "0/2" in result.output


def test_lists_cases(cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers):
    """Test to that cases displays case in database"""

    # GIVEN a database with a case
    base_store: Store = base_context.status_db
    case = helpers.add_case(base_store)

    # WHEN listing cases
    result = cli_runner.invoke(status_of_cases, obj=base_context)

    # THEN the case should be listed
    assert result.exit_code == 0
    assert case.internal_id in result.output



"""Fixtures for cli workflow rnafusion tests"""

import datetime as dt
import gzip
from pathlib import Path
from typing import List

import pytest

from cg.apps.hermes.hermes_api import HermesApi
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import Pipeline
from cg.constants.constants import FileFormat
from cg.io.controller import WriteFile, WriteStream
from cg.io.json import read_json, write_json
from cg.meta.workflow.rnafusion import RnafusionAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Family, Sample
from tests.mocks.process_mock import ProcessMock
from tests.mocks.tb_mock import MockTB
from tests.store_helpers import StoreHelpers


@pytest.fixture(name="rnafusion_dir")
def rnafusion_dir(tmpdir_factory, apps_dir: Path) -> str:
    """Return the path to the rnafusion apps dir."""
    rnafusion_dir = tmpdir_factory.mktemp("rnafusion")
    return Path(rnafusion_dir).absolute().as_posix()


@pytest.fixture(name="rnafusion_case_id")
def fixture_rnafusion_case_id() -> str:
    """Returns a rnafusion case id."""
    return "rnafusion_case_enough_reads"


@pytest.fixture(name="no_sample_case_id")
def fixture_no_sample_case_id() -> str:
    """Returns a case id of a case with no samples."""
    return "no_sample_case"


@pytest.fixture(name="rnafusion_sample_id")
def fixture_rnafusion_sample_id() -> str:
    """Returns a rnafusion sample id."""
    return "sample_rnafusion_case_enough_reads"


@pytest.fixture(name="rnafusion_housekeeper_dir")
def rnafusion_housekeeper_dir(tmpdir_factory, rnafusion_dir: Path) -> Path:
    """Return the path to the rnafusion housekeeper bundle dir."""
    return tmpdir_factory.mktemp("bundles")


@pytest.fixture(name="rnafusion_reference_path")
def rnafusion_reference_path(rnafusion_dir: Path) -> str:
    rnafusion_reference_path = Path(rnafusion_dir, "references")
    rnafusion_reference_path.touch(exist_ok=True)
    return rnafusion_reference_path.as_posix()


@pytest.fixture
def rnafusion_fastq_file_l_1_r_1(rnafusion_housekeeper_dir: Path) -> str:
    fastq_filename = Path(
        rnafusion_housekeeper_dir, "XXXXXXXXX_000000_S000_L001_R1_001.fastq.gz"
    ).as_posix()
    with gzip.open(fastq_filename, "wb") as wh:
        wh.write(b"@A00689:73:XXXXXXXXX:1:1101:4806:1047 1:N:0:TCCTGGAACA+ACAACCAGTA")
    return fastq_filename


@pytest.fixture
def rnafusion_fastq_file_l_1_r_2(rnafusion_housekeeper_dir: Path) -> str:
    fastq_filename = Path(
        rnafusion_housekeeper_dir, "XXXXXXXXX_000000_S000_L001_R2_001.fastq.gz"
    ).as_posix()
    with gzip.open(fastq_filename, "wb") as wh:
        wh.write(b"@A00689:73:XXXXXXXXX:1:1101:4806:1047 2:N:0:TCCTGGAACA+ACAACCAGTA")
    return fastq_filename


@pytest.fixture
def rnafusion_mock_fastq_files(
    rnafusion_fastq_file_l_1_r_1: Path, rnafusion_fastq_file_l_1_r_2: Path
) -> List[Path]:
    """Return list of all mock fastq files to commit to mock housekeeper"""
    return [rnafusion_fastq_file_l_1_r_1, rnafusion_fastq_file_l_1_r_2]


@pytest.fixture(scope="function", name="rnafusion_housekeeper")
def rnafusion_housekeeper(
    housekeeper_api: HousekeeperAPI,
    helpers: StoreHelpers,
    rnafusion_mock_fastq_files: List[Path],
    rnafusion_sample_id: str,
):
    """Create populated housekeeper that holds files for all mock samples."""

    bundle_data = {
        "name": rnafusion_sample_id,
        "created": dt.datetime.now(),
        "version": "1.0",
        "files": [
            {"path": f, "tags": ["fastq"], "archive": False} for f in rnafusion_mock_fastq_files
        ],
    }
    helpers.ensure_hk_bundle(store=housekeeper_api, bundle_data=bundle_data)
    return housekeeper_api


@pytest.fixture(scope="function", name="rnafusion_context")
def fixture_rnafusion_context(
    cg_context: CGConfig,
    helpers: StoreHelpers,
    rnafusion_housekeeper: HousekeeperAPI,
    trailblazer_api: MockTB,
    hermes_api: HermesApi,
    cg_dir: Path,
    rnafusion_case_id: str,
    rnafusion_sample_id: str,
    no_sample_case_id: str,
) -> CGConfig:
    """context to use in cli"""
    cg_context.housekeeper_api_ = rnafusion_housekeeper
    cg_context.trailblazer_api_ = trailblazer_api
    cg_context.meta_apis["analysis_api"] = RnafusionAnalysisAPI(config=cg_context)
    status_db: Store = cg_context.status_db

    # Create ERROR case with NO SAMPLES
    helpers.add_case(status_db, internal_id=no_sample_case_id, name=no_sample_case_id)

    # Create textbook case with enough reads
    case_enough_reads: Family = helpers.add_case(
        store=status_db,
        internal_id=rnafusion_case_id,
        name=rnafusion_case_id,
        data_analysis=Pipeline.RNAFUSION,
    )

    sample_rnafusion_case_enough_reads: Sample = helpers.add_sample(
        status_db,
        internal_id=rnafusion_sample_id,
        sequenced_at=dt.datetime.now(),
    )

    helpers.add_relationship(
        status_db,
        case=case_enough_reads,
        sample=sample_rnafusion_case_enough_reads,
    )
    return cg_context


@pytest.fixture(name="deliverable_data")
def fixture_deliverables_data(
    rnafusion_dir: Path, rnafusion_case_id: str, rnafusion_sample_id: str
) -> dict:
    return {
        "files": [
            {
                "path": f"{rnafusion_dir}/{rnafusion_case_id}/multiqc/multiqc_report.html",
                "path_index": "",
                "step": "report",
                "tag": ["multiqc-html", "rna"],
                "id": rnafusion_case_id,
                "format": "html",
                "mandatory": True,
            },
        ]
    }


@pytest.fixture
def mock_deliverable(rnafusion_dir: Path, deliverable_data: dict, rnafusion_case_id: str) -> None:
    """Create deliverable file with dummy data and files to deliver."""
    Path.mkdir(
        Path(rnafusion_dir, rnafusion_case_id),
        parents=True,
        exist_ok=True,
    )
    Path.mkdir(
        Path(rnafusion_dir, rnafusion_case_id, "multiqc"),
        parents=True,
        exist_ok=True,
    )
    for report_entry in deliverable_data["files"]:
        Path(report_entry["path"]).touch(exist_ok=True)
    WriteFile.write_file_from_content(
        content=deliverable_data,
        file_format=FileFormat.JSON,
        file_path=Path(rnafusion_dir, rnafusion_case_id, rnafusion_case_id + "_deliverables.yaml"),
    )


@pytest.fixture(name="hermes_deliverables")
def fixture_hermes_deliverables(deliverable_data: dict, rnafusion_case_id: str) -> dict:
    hermes_output: dict = {"pipeline": "rnafusion", "bundle_id": rnafusion_case_id, "files": []}
    for file_info in deliverable_data["files"]:
        tags: List[str] = []
        if "html" in file_info["format"]:
            tags.append("multiqc-html")
        hermes_output["files"].append({"path": file_info["path"], "tags": tags, "mandatory": True})
    return hermes_output


@pytest.fixture(name="malformed_hermes_deliverables")
def fixture_malformed_hermes_deliverables(hermes_deliverables: dict) -> dict:
    malformed_deliverable: dict = hermes_deliverables.copy()
    malformed_deliverable.pop("pipeline")

    return malformed_deliverable


@pytest.fixture(name="rnafusion_hermes_process")
def fixture_rnafusion_hermes_process(
    hermes_deliverables: dict, process: ProcessMock
) -> ProcessMock:
    """Return a process mock populated with some rnafusion hermes output."""
    process.set_stdout(
        text=WriteStream.write_stream_from_content(
            content=hermes_deliverables, file_format=FileFormat.JSON
        )
    )
    return process


@pytest.fixture(name="rnafusion_multiqc_json_metrics")
def fixture_rnafusion_multiqc_json_metrics(rnafusion_analysis_dir) -> dict:
    """Returns a the content of a mock multiqc json file."""
    return read_json(file_path=Path(rnafusion_analysis_dir, "multiqc_data.json"))


@pytest.fixture
def mock_analysis_finish(
    rnafusion_dir: Path, rnafusion_case_id: str, rnafusion_multiqc_json_metrics: dict
) -> None:
    """Create analysis_finish file for testing."""
    Path.mkdir(Path(rnafusion_dir, rnafusion_case_id, "pipeline_info"), parents=True, exist_ok=True)
    Path(rnafusion_dir, rnafusion_case_id, "pipeline_info", "software_versions.yml").touch(
        exist_ok=True
    )
    Path(rnafusion_dir, rnafusion_case_id, f"{rnafusion_case_id}_samplesheet.csv").touch(
        exist_ok=True
    )
    Path.mkdir(
        Path(rnafusion_dir, rnafusion_case_id, "multiqc", "multiqc_data"),
        parents=True,
        exist_ok=True,
    )
    write_json(
        content=rnafusion_multiqc_json_metrics,
        file_path=Path(
            rnafusion_dir,
            rnafusion_case_id,
            "multiqc",
            "multiqc_data",
            "multiqc_data.json",
        ),
    )


@pytest.fixture
def mock_config(rnafusion_dir: Path, rnafusion_case_id: str) -> None:
    """Create samplesheet.csv file for testing"""
    Path.mkdir(Path(rnafusion_dir, rnafusion_case_id), parents=True, exist_ok=True)
    Path(rnafusion_dir, rnafusion_case_id, f"{rnafusion_case_id}_samplesheet.csv").touch(
        exist_ok=True
    )



import logging
from pathlib import Path

import pytest
from cg.constants.demultiplexing import FlowCellType, UNKNOWN_REAGENT_KIT_VERSION
from cg.exc import FlowCellError
from cg.models.demultiplex.run_parameters import RunParameters


def test_get_flow_cell_type_hiseq_run_parameters(hiseq_run_parameters: Path):
    """Test that reading a hiseq run_parameters file assigns the correct flow cell type."""
    # GIVEN a RunParameters object created from a Hiseq run parameters file
    run_parameters = RunParameters(run_parameters_path=hiseq_run_parameters)

    # WHEN fetching the flow cell type

    # THEN the flow cell type is hiseq
    assert run_parameters.flow_cell_type == FlowCellType.HISEQ


def test_get_flow_cell_type_novaseq_run_parameters(novaseq_run_parameters: Path):
    """Test that reading a Novaseq run_parameters file assigns the correct flow cell type."""
    # GIVEN a RunParameters object created from a Novaseq run parameters file
    run_parameters = RunParameters(run_parameters_path=novaseq_run_parameters)

    # WHEN fetching the flow cell type

    # THEN assert that the flow cell type is novaseq
    assert run_parameters.flow_cell_type == FlowCellType.NOVASEQ


def test_get_unknown_type_run_parameters(unknown_run_parameters: Path, caplog):
    """Test that reading a run_parameters with unknown flow cell file raises an error."""
    caplog.set_level(logging.INFO)
    # GIVEN a RunParameters object created from a run parameters file with unknown flow cell
    run_parameters = RunParameters(run_parameters_path=unknown_run_parameters)

    # WHEN fetching the flow cell type
    with pytest.raises(FlowCellError):
        # THEN assert that an exception was raised since the flow cell type was unknown
        run_parameters.flow_cell_type

    assert "Unknown flow cell type" in caplog.text


def test_get_flow_cell_type_when_missing(run_parameters_missing_flowcell_type: Path, caplog):
    """Test that reading a run_parameters with missing flow cell file raises an error."""
    caplog.set_level(logging.INFO)
    # GIVEN a RunParameters object created from a run parameters file with missing flow cell
    run_parameters = RunParameters(run_parameters_path=run_parameters_missing_flowcell_type)

    # WHEN parsing the run parameters
    with pytest.raises(FlowCellError):
        # THEN assert that an exception was raised since the flow cell type was unknown
        run_parameters.flow_cell_type

    assert "Could not determine flow cell type" in caplog.text


def test_get_base_mask(novaseq_run_parameters: Path):
    """Test that generating the run parameters base mask returns the expected result."""
    # GIVEN a RunParameters object created from a valid run parameters file
    run_parameters = RunParameters(run_parameters_path=novaseq_run_parameters)

    # WHEN fetching the base mask
    base_mask: str = run_parameters.get_base_mask()

    # THEN the base mask has the correct structure
    assert base_mask.startswith("Y")
    assert ",I" in base_mask


def test_get_base_mask_undetermined_cycles(run_parameters_missing_flowcell_type: Path, caplog):
    """Test that generating the base mask from a file without cycle info fails."""
    caplog.set_level(logging.INFO)
    # GIVEN a RunParameters object created from a run parameters file with missing cycle info
    run_parameters = RunParameters(run_parameters_path=run_parameters_missing_flowcell_type)

    # WHEN creating the base mask
    with pytest.raises(FlowCellError):
        # THEN assert that an exception was raised since the cycle data was not found
        run_parameters.get_base_mask()

    assert "Could not determine length of reads one" in caplog.text


def test_flow_cell_mode(novaseq_run_parameters: Path):
    """Test that the flow cell mode is of a valid file in the list of expected values."""
    # GIVEN a RunParameters object created from a valid run parameters file
    run_parameters = RunParameters(run_parameters_path=novaseq_run_parameters)

    # WHEN fetching the flow cell mode

    # THEN the flow cell mode is in the list of expected values
    assert run_parameters.flow_cell_mode in ["S2", "S4"]


def test_flow_cell_mode_missing_mode(run_parameters_missing_flowcell_type: Path):
    """Test that getting the flow cell mode from a file lacking flow cell mode returns None."""
    # GIVEN a RunParameters object created from a file without flow cell mode
    run_parameters = RunParameters(run_parameters_path=run_parameters_missing_flowcell_type)

    # WHEN fetching the flow cell mode

    # THEN the flow cell mode is none
    assert run_parameters.flow_cell_mode is None


def test_reagent_kit_version(novaseq_run_parameters: Path):
    """Test that getting reagent kit version from a correct file returns an expected value."""
    # GIVEN a RunParameters object created from a valid run parameters file
    run_parameters = RunParameters(run_parameters_path=novaseq_run_parameters)

    # WHEN fetching the reagent kit version
    reagent_kit_version: str = run_parameters.reagent_kit_version

    # THEN the reagent kit version exists and is not "unknown"
    assert reagent_kit_version
    assert reagent_kit_version != UNKNOWN_REAGENT_KIT_VERSION


def test_reagent_kit_version_missing_version(run_parameters_missing_flowcell_type: Path, caplog):
    """Test that 'unknown' will be returned if the run parameters file has no reagent kit method."""
    caplog.set_level(logging.INFO)
    # GIVEN a RunParameters object created from a file without reagent kit version
    run_parameters = RunParameters(run_parameters_path=run_parameters_missing_flowcell_type)

    # WHEN fetching the reagent kit version

    # THEN the reagent kit version is unknown
    assert run_parameters.reagent_kit_version == UNKNOWN_REAGENT_KIT_VERSION
    assert "Could not determine reagent kit version" in caplog.text


def test_control_software_version(novaseq_run_parameters: Path):
    """Test that getting control software version from a correct file returns an expected value."""
    # GIVEN a RunParameters object created from a valid run parameters file
    run_parameters = RunParameters(run_parameters_path=novaseq_run_parameters)

    # WHEN fetching the control software version

    # THEN the control software version is a string
    assert isinstance(run_parameters.control_software_version, str)


def test_control_software_version_no_version(run_parameters_missing_flowcell_type: Path, caplog):
    """Test that fetching the control software version from a file without that field fails."""
    caplog.set_level(logging.INFO)
    # GIVEN a RunParameters object created from a file without control software version
    run_parameters = RunParameters(run_parameters_path=run_parameters_missing_flowcell_type)

    # WHEN fetching the control software version
    with pytest.raises(FlowCellError):
        # THEN assert that an exception was raised since the control software version was not found
        run_parameters.control_software_version

    assert "Could not determine control software version" in caplog.text


def test_index_length(novaseq_run_parameters: Path):
    """Test that getting the index length from a valid file returns the expected output."""
    # GIVEN a RunParameters object created from a valid run parameters file
    run_parameters = RunParameters(run_parameters_path=novaseq_run_parameters)

    # WHEN getting the index length

    # THEN the index length is an int
    assert isinstance(run_parameters.index_length, int)


def test_index_length_different_length(run_parameters_different_index: Path):
    """Test that getting the index length from a file with different index cycles fails."""
    # GIVEN a RunParameters object created from a file with different index cycles
    run_parameters = RunParameters(run_parameters_path=run_parameters_different_index)

    # WHEN fetching index length
    with pytest.raises(FlowCellError) as exc_info:
        # THEN assert that an exception was raised since the index cycles are different
        run_parameters.index_length

    assert str(exc_info.value) == "Index lengths are not the same!"



"""Tests for MicroSALT analysis."""
from typing import List

from cg.apps.tb import TrailblazerAPI
from cg.meta.workflow.microsalt import MicrosaltAnalysisAPI
from cg.models.cg_config import CGConfig
from pathlib import Path
import logging

from cg.models.orders.sample_base import ControlEnum
from cg.store import Store
from cg.store.models import Family


def test_qc_check_fail(
    qc_microsalt_context: CGConfig,
    microsalt_qc_fail_run_dir_path: Path,
    microsalt_qc_fail_lims_project: str,
    microsalt_case_qc_fail: str,
    caplog,
    mocker,
):
    """QC check for a microsalt case that should fail."""
    caplog.set_level(logging.INFO)
    store: Store = qc_microsalt_context.status_db
    microsalt_api: MicrosaltAnalysisAPI = qc_microsalt_context.meta_apis["analysis_api"]

    # GIVEN a case that is to be stored
    microsalt_case: Family = store.get_case_by_internal_id(internal_id=microsalt_case_qc_fail)
    for index in range(4):
        microsalt_case.samples[index].reads = 1000

    mocker.patch.object(MicrosaltAnalysisAPI, "create_qc_done_file")

    # WHEN performing QC check
    qc_pass: bool = microsalt_api.microsalt_qc(
        case_id=microsalt_case_qc_fail,
        run_dir_path=microsalt_qc_fail_run_dir_path,
        lims_project=microsalt_qc_fail_lims_project,
    )

    # THEN the QC should fail
    assert not qc_pass
    assert "failed" in caplog.text


def test_qc_check_pass(
    qc_microsalt_context: CGConfig,
    microsalt_qc_pass_run_dir_path: Path,
    microsalt_qc_pass_lims_project: str,
    microsalt_case_qc_pass: str,
    caplog,
    mocker,
):
    """QC check for a microsalt case that should pass."""
    caplog.set_level(logging.INFO)
    store: Store = qc_microsalt_context.status_db
    microsalt_api: MicrosaltAnalysisAPI = qc_microsalt_context.meta_apis["analysis_api"]

    # GIVEN a case that is to be stored
    microsalt_case: Family = store.get_case_by_internal_id(internal_id=microsalt_case_qc_pass)
    microsalt_case.samples[1].control = ControlEnum.negative
    microsalt_case.samples[1].reads = 1100000

    mocker.patch.object(MicrosaltAnalysisAPI, "create_qc_done_file")

    # WHEN performing QC check
    qc_pass: bool = microsalt_api.microsalt_qc(
        case_id=microsalt_case_qc_pass,
        run_dir_path=microsalt_qc_pass_run_dir_path,
        lims_project=microsalt_qc_pass_lims_project,
    )

    # THEN the QC should pass
    assert qc_pass
    assert "passed" in caplog.text


def test_qc_check_negative_control_fail(
    qc_microsalt_context: CGConfig,
    microsalt_qc_fail_run_dir_path: Path,
    microsalt_qc_fail_lims_project: str,
    microsalt_case_qc_fail: str,
    caplog,
    mocker,
):
    """QC check for a microsalt case where a negative control fails QC."""

    caplog.set_level(logging.INFO)
    store = qc_microsalt_context.status_db
    microsalt_api: MicrosaltAnalysisAPI = qc_microsalt_context.meta_apis["analysis_api"]

    # GIVEN a case that is to be stored
    microsalt_case: Family = store.get_case_by_internal_id(internal_id=microsalt_case_qc_fail)
    microsalt_case.samples[0].control = ControlEnum.negative

    mocker.patch.object(MicrosaltAnalysisAPI, "create_qc_done_file")

    # WHEN performing QC check
    qc_pass: bool = microsalt_api.microsalt_qc(
        case_id=microsalt_case_qc_fail,
        run_dir_path=microsalt_qc_fail_run_dir_path,
        lims_project=microsalt_qc_fail_lims_project,
    )

    # THEN the QC should fail
    assert not qc_pass
    assert "failed" in caplog.text
    assert "Negative control sample" in caplog.text


def test_get_cases_to_store_pass(
    qc_microsalt_context: CGConfig,
    caplog,
    mocker,
    microsalt_qc_pass_lims_project: str,
    microsalt_case_qc_pass: str,
    microsalt_qc_pass_run_dir_path: Path,
):
    """Test get cases to store for a microsalt case that passes QC."""

    caplog.set_level(logging.INFO)
    store = qc_microsalt_context.status_db
    microsalt_api: MicrosaltAnalysisAPI = qc_microsalt_context.meta_apis["analysis_api"]
    mocker.patch.object(MicrosaltAnalysisAPI, "create_qc_done_file")
    mocker.patch.object(TrailblazerAPI, "set_analysis_status")
    mocker.patch.object(TrailblazerAPI, "add_comment")

    # GIVEN a store with a QC ready microsalt case that will pass QC
    microsalt_pass_case: Family = store.get_case_by_internal_id(internal_id=microsalt_case_qc_pass)
    microsalt_pass_case.samples[1].control = "negative"
    microsalt_pass_case.samples[1].reads = 1100000

    mocker.patch.object(
        MicrosaltAnalysisAPI,
        "get_completed_cases",
        return_value=[microsalt_pass_case],
    )
    mocker.patch.object(
        MicrosaltAnalysisAPI, "get_project", return_value=microsalt_qc_pass_lims_project
    )

    mocker.patch.object(
        MicrosaltAnalysisAPI, "get_latest_case_path", return_value=microsalt_qc_pass_run_dir_path
    )

    # WHEN get cases to store
    cases_to_store: List[Family] = microsalt_api.get_cases_to_store()

    # THEN it should be stored
    assert microsalt_pass_case in cases_to_store


def test_get_cases_to_store_fail(
    qc_microsalt_context: CGConfig,
    caplog,
    mocker,
    microsalt_qc_fail_lims_project: str,
    microsalt_case_qc_fail: str,
    microsalt_qc_fail_run_dir_path: Path,
):
    """Test get cases to store for a microsalt case that fails QC."""

    caplog.set_level(logging.INFO)
    store = qc_microsalt_context.status_db
    microsalt_api: MicrosaltAnalysisAPI = qc_microsalt_context.meta_apis["analysis_api"]
    mocker.patch.object(MicrosaltAnalysisAPI, "create_qc_done_file")
    mocker.patch.object(TrailblazerAPI, "set_analysis_status")
    mocker.patch.object(TrailblazerAPI, "add_comment")

    # GIVEN a store with a QC ready microsalt case that will fail QC
    microsalt_fail_case: Family = store.get_case_by_internal_id(internal_id=microsalt_case_qc_fail)

    mocker.patch.object(
        MicrosaltAnalysisAPI,
        "get_completed_cases",
        return_value=[microsalt_fail_case],
    )
    mocker.patch.object(
        MicrosaltAnalysisAPI, "get_project", return_value=microsalt_qc_fail_lims_project
    )

    mocker.patch.object(
        MicrosaltAnalysisAPI, "get_latest_case_path", return_value=microsalt_qc_fail_run_dir_path
    )

    # WHEN get case to store
    cases_to_store: List[Family] = microsalt_api.get_cases_to_store()

    # Then it should not be stored
    assert microsalt_fail_case not in cases_to_store


def test_get_latest_case_path(
    mocker,
    qc_microsalt_context: CGConfig,
    microsalt_case_qc_pass: str,
    microsalt_analysis_dir: Path,
):
    """Test get_latest_case_path return the first case path and not single sample path"""
    microsalt_api: MicrosaltAnalysisAPI = qc_microsalt_context.meta_apis["analysis_api"]

    # GIVEN a case with different case paths, both single sample and case analyses
    mocker.patch.object(MicrosaltAnalysisAPI, "get_project", return_value="ACC12345")
    mocker.patch.object(
        MicrosaltAnalysisAPI,
        "get_case_path",
        return_value=[
            Path(microsalt_analysis_dir, "ACC12345A2_2023"),
            Path(microsalt_analysis_dir, "ACC12345_2022"),
            Path(microsalt_analysis_dir, "ACC12345A1_2023"),
        ],
    )
    # WHEN getting the latest case path
    path = microsalt_api.get_latest_case_path(case_id=microsalt_case_qc_pass)

    # THEN the first case path should be returned
    assert Path(microsalt_analysis_dir, "ACC12345_2022") == path



import logging
from pathlib import Path
from typing import Optional
from xml.etree import ElementTree

from cg.constants.demultiplexing import FlowCellType, UNKNOWN_REAGENT_KIT_VERSION
from cg.exc import FlowCellError
from typing_extensions import Literal

LOG = logging.getLogger(__name__)


class RunParameters:
    """Class to handle the run parameters from a sequencing run."""

    def __init__(self, run_parameters_path: Path):
        self.path: Path = run_parameters_path
        with open(run_parameters_path, "rt") as in_file:
            self.tree: ElementTree = ElementTree.parse(in_file)

    @property
    def index_length(self) -> int:
        """Return the length of the indexes if they are equal, raise an error otherwise."""
        index_one_length: int = self.get_index1_cycles()
        index_two_length: int = self.get_index2_cycles()
        if index_one_length != index_two_length:
            raise FlowCellError("Index lengths are not the same!")
        return index_one_length

    @property
    def control_software_version(self) -> str:
        """Return the control software version."""
        node_name: str = ".ApplicationVersion"
        xml_node: Optional[ElementTree.Element] = self.tree.find(node_name)
        self.node_not_found(node=xml_node, name="control software version")
        return xml_node.text

    @property
    def reagent_kit_version(self) -> str:
        """Return the reagent kit version if existent, return 'unknown' otherwise."""
        node_name: str = "./RfidsInfo/SbsConsumableVersion"
        xml_node: Optional[ElementTree.Element] = self.tree.find(node_name)
        if xml_node is None:
            LOG.warning("Could not determine reagent kit version")
            LOG.info("Set reagent kit version to 'unknown'")
            return UNKNOWN_REAGENT_KIT_VERSION
        return xml_node.text

    @property
    def flow_cell_type(self) -> Literal[FlowCellType.NOVASEQ, FlowCellType.HISEQ]:
        """Fetch the flow cell type from the run parameters."""
        # First try with the node name for hiseq
        node_name: str = "./Setup/ApplicationName"
        xml_node: Optional[ElementTree.Element] = self.tree.find(node_name)
        if xml_node is None:
            # Then try with node name for novaseq
            node_name: str = ".Application"
            xml_node: Optional[ElementTree.Element] = self.tree.find(node_name)
        self.node_not_found(node=xml_node, name="flow cell type")
        for flow_cell_name in [FlowCellType.NOVASEQ, FlowCellType.HISEQ]:
            if flow_cell_name in xml_node.text.lower():
                return flow_cell_name
        message = f"Unknown flow cell type {xml_node.text}"
        LOG.warning(message)
        raise FlowCellError(message)

    @property
    def flow_cell_mode(self) -> Optional[str]:
        """Return the flow cell mode."""
        node_name: str = "/RfidsInfo/FlowCellMode"
        xml_node: Optional[ElementTree.Element] = self.tree.find(node_name)
        if xml_node is None:
            LOG.warning("Could not determine flow cell mode")
            LOG.info("Set flow cell mode to None")
            return
        return xml_node.text

    @property
    def requires_dummy_samples(self) -> bool:
        """Return true if the flow cell requires the addition of dummy samples.

        If the number of cycles of both indexes is 8, the flow cell does not need the addition of dummy samples.
        """
        return self.index_length != 8

    @staticmethod
    def node_not_found(node: Optional[ElementTree.Element], name: str) -> None:
        """Raise exception if the given node is not found."""
        if node is None:
            message = f"Could not determine {name}"
            LOG.warning(message)
            raise FlowCellError(message)

    def get_node_integer_value(self, node_name: str, name: str) -> int:
        """Return the value of the node as an integer."""
        xml_node = self.tree.find(node_name)
        self.node_not_found(node=xml_node, name=name)
        return int(xml_node.text)

    def get_index1_cycles(self) -> int:
        """Return the number of cycles in the first index read."""
        node_name = "./IndexRead1NumberOfCycles"
        return self.get_node_integer_value(node_name=node_name, name="length of index one")

    def get_index2_cycles(self) -> int:
        """Return the number of cycles in the second index read."""
        node_name = "./IndexRead2NumberOfCycles"
        return self.get_node_integer_value(node_name=node_name, name="length of index two")

    def get_read1_cycles(self) -> int:
        """Return the number of cycles in the first read."""
        node_name = "./Read1NumberOfCycles"
        return self.get_node_integer_value(node_name=node_name, name="length of reads one")

    def get_read2_cycles(self) -> int:
        """Return the number of cycles in the second read."""
        node_name = "./Read2NumberOfCycles"
        return self.get_node_integer_value(node_name=node_name, name="length of reads two")

    def get_base_mask(self) -> str:
        """Create the basemask for novaseq flow cells.

        Basemask is used in this comma format as an argument to bcl2fastq.
        When creating the unaligned path the commas are stripped.
        """
        return (
            f"Y{self.get_read1_cycles()},"
            f"I{self.get_index1_cycles()},"
            f"I{self.get_index2_cycles()},"
            f"Y{self.get_read2_cycles()}"
        )

    def __str__(self):
        return (
            f"RunParameters(path={self.path},"
            f"flow_cell_type={self.flow_cell_type},"
            f"flow_cell_mode={self.flow_cell_mode})"
        )

    def __repr__(self):
        return (
            f"RunParameters(path={self.path},flow_cell_type={self.flow_cell_type},flow_cell_mode={self.flow_cell_mode},"
            f"reagent_kit_version={self.reagent_kit_version},control_software_version={self.control_software_version},"
            f"index_length={self.index_length})"
        )



import logging
from typing import Dict, Iterable, Optional, Union

import sqlalchemy

from cg.apps.cgstats.db.models import (
    Datasource,
    Demux,
    DemuxSample,
    DragenDemuxSample,
    Flowcell,
    Project,
    Supportparams,
    Sample,
    Unaligned,
)
from cg.apps.cgstats.stats import StatsAPI
from cg.apps.demultiplex.sample_sheet.models import NovaSeqSample, SampleSheet
from cg.constants.demultiplexing import DRAGEN_PASSED_FILTER_PCT
from cg.constants.symbols import PERIOD
from cg.models.demultiplex.demux_results import DemuxResults, LogfileParameters

LOG = logging.getLogger(__name__)


def create_support_parameters(manager: StatsAPI, demux_results: DemuxResults) -> Supportparams:
    logfile_parameters: LogfileParameters = demux_results.get_logfile_parameters()
    support_parameters = manager.Supportparams()
    support_parameters.document_path = str(
        demux_results.results_dir
    )  # This is the unaligned directory
    support_parameters.idstring = logfile_parameters.id_string
    support_parameters.program = logfile_parameters.program
    support_parameters.commandline = logfile_parameters.command_line
    support_parameters.sampleconfig_path = str(demux_results.sample_sheet_path)
    support_parameters.sampleconfig = demux_results.sample_sheet_path.read_text()
    support_parameters.time = logfile_parameters.time
    manager.add(support_parameters)
    manager.flush()
    LOG.info("Creating new support parameters object %s", support_parameters)
    return support_parameters


def create_datasource(
    manager: StatsAPI, demux_results: DemuxResults, support_parameters_id: int
) -> Datasource:
    datasource = manager.Datasource()
    datasource.runname = demux_results.run_name
    datasource.rundate = demux_results.run_date
    datasource.machine = demux_results.machine_name
    datasource.server = demux_results.demux_host
    datasource.document_path = str(demux_results.conversion_stats_path)
    datasource.document_type = demux_results.conversion_stats_path.suffix.strip(PERIOD)
    datasource.time = sqlalchemy.func.now()
    datasource.supportparams_id = support_parameters_id

    manager.add(datasource)
    manager.flush()
    LOG.info("Creating new datasource object %s", datasource)
    return datasource


def create_flowcell(manager: StatsAPI, demux_results: DemuxResults) -> Flowcell:
    flowcell = manager.Flowcell()
    flowcell.flowcellname = demux_results.flow_cell.id
    flowcell.flowcell_pos = demux_results.flow_cell.position
    flowcell.hiseqtype = "novaseq"
    flowcell.time = sqlalchemy.func.now()

    manager.add(flowcell)
    manager.flush()
    LOG.info("Creating new flowcell object %s", flowcell)
    return flowcell


def create_demux(
    manager: StatsAPI,
    datasource_id: int,
    demux_results: DemuxResults,
    flow_cell_id: int,
) -> Demux:
    demux: Demux = manager.Demux()
    demux.flowcell_id = flow_cell_id
    demux.datasource_id = datasource_id
    if demux_results.bcl_converter == "dragen":
        demux.basemask: str = demux_results.run_info.basemask
    else:
        demux.basemask = ""
    demux.time = sqlalchemy.func.now()

    manager.add(demux)
    manager.flush()
    LOG.info("Creating new demux object %s", demux)
    return demux


def create_project(manager: StatsAPI, project_name: str) -> Project:
    project: Project = manager.Project()
    project.projectname = project_name
    project.time = sqlalchemy.func.now()
    manager.add(project)
    manager.flush()
    LOG.info("Creating new project object %s", project)
    return project


def create_sample(manager: StatsAPI, sample_id: str, barcode: str, project_id: int) -> Sample:
    sample: Sample = manager.Sample()
    sample.project_id = project_id
    sample.samplename = sample_id
    sample.limsid = sample_id.split("_")[0]
    sample.barcode = barcode
    sample.time = sqlalchemy.func.now()

    manager.add(sample)
    manager.flush()
    return sample


def create_unaligned(
    manager: StatsAPI, demux_sample: DemuxSample, sample_id: int, demux_id: int
) -> Unaligned:
    unaligned: Unaligned = manager.Unaligned()
    unaligned.sample_id = sample_id
    unaligned.demux_id = demux_id
    unaligned.lane = demux_sample.lane
    unaligned.yield_mb = round(int(demux_sample.pass_filter_yield) / 1000000, 2)
    unaligned.passed_filter_pct = demux_sample.pass_filter_yield_pc
    unaligned.readcounts = demux_sample.pass_filter_clusters * 2
    unaligned.raw_clusters_per_lane_pct = demux_sample.raw_clusters_pc
    unaligned.perfect_indexreads_pct = (
        round(demux_sample.perfect_barcodes / demux_sample.barcodes * 100, 5)
        if demux_sample.barcodes
        else 0
    )
    unaligned.q30_bases_pct = demux_sample.pass_filter_Q30
    unaligned.mean_quality_score = demux_sample.pass_filter_qscore
    unaligned.time = sqlalchemy.func.now()

    manager.add(unaligned)
    manager.flush()
    return unaligned


def create_dragen_unaligned(
    manager: StatsAPI, demux_sample: DragenDemuxSample, sample_id: int, demux_id: int
) -> Unaligned:
    """Create an unaligned object in cgstats for a sample demultiplexed with Dragen"""
    unaligned: Unaligned = manager.Unaligned()
    unaligned.sample_id: int = sample_id
    unaligned.demux_id: int = demux_id
    unaligned.lane: int = demux_sample.lane
    unaligned.passed_filter_pct: float = DRAGEN_PASSED_FILTER_PCT
    unaligned.readcounts: int = _calculate_read_counts(demux_sample)
    unaligned.perfect_indexreads_pct: float = _calculate_perfect_indexreads_pct(demux_sample)
    unaligned.q30_bases_pct: float = _calculate_q30_bases_pct(demux_sample)
    unaligned.yield_mb: float = _calculate_yield(demux_sample)
    unaligned.mean_quality_score: float = demux_sample.mean_quality_score
    unaligned.time: sqlalchemy.sql.func.now = sqlalchemy.func.now()

    manager.add(unaligned)
    manager.flush()
    return unaligned


def _calculate_perfect_indexreads_pct(demux_sample: DragenDemuxSample) -> float:
    """calculates the percentage of perfect index reads"""
    return (
        round(demux_sample.perfect_reads / demux_sample.reads * 100, 2) if demux_sample.reads else 0
    )


def _calculate_q30_bases_pct(demux_sample: DragenDemuxSample) -> float:
    """calculates the percentage of bases with a sequencing quality score of 30 or over"""
    return (
        round(
            demux_sample.pass_filter_q30
            / (demux_sample.r1_sample_bases + demux_sample.r2_sample_bases)
            * 100,
            2,
        )
        if demux_sample.r1_sample_bases + demux_sample.r2_sample_bases
        else 0
    )


def _calculate_yield(demux_sample: DragenDemuxSample) -> float:
    """calculates the amount of data produced in MB"""
    total_reads = _calculate_read_counts(demux_sample)
    return round(total_reads * demux_sample.read_length / 1000000, 0)


def _calculate_read_counts(demux_sample: DragenDemuxSample) -> int:
    """calculates the number of reads from the number of clusters"""
    return demux_sample.reads * 2


def create_projects(manager: StatsAPI, project_names: Iterable[str]) -> Dict[str, int]:
    project_name_to_id: Dict[str, int] = {}
    for project_name in project_names:
        project: Optional[Project] = get_or_create_project(
            manager=manager, project_name=project_name
        )
        project_name_to_id[project_name] = project.project_id
    return project_name_to_id


def get_or_create_sample(
    manager: StatsAPI, sample: NovaSeqSample, project_name_to_id: Dict[str, int]
) -> Sample:
    """Create a new Sample object in the cgstats database if it doesn't already exist."""

    if sample.project == "indexcheck":
        LOG.debug("Skip adding indexcheck sample to database")
        return None

    barcode = f"{sample.index}+{sample.second_index}" if sample.second_index else sample.index

    stats_sample: Optional[Sample] = manager.find_handler.get_sample_by_name_and_barcode(
        sample_name=sample.sample_id, barcode=barcode
    )

    if not stats_sample:
        project_id = project_name_to_id[sample.project]

        stats_sample: Sample = create_sample(
            manager=manager,
            sample_id=sample.sample_id,
            barcode=barcode,
            project_id=project_id,
        )

    return stats_sample


def _create_dragen_samples(
    manager: StatsAPI,
    demux_results: DemuxResults,
    project_name_to_id: Dict[str, int],
    demux_id: int,
    sample_sheet: SampleSheet,
):
    """Handles sample creation: creates sample objects and unaligned objects in their respective
    tables in cgstats for samples demultiplexed with Dragen"""

    demux_samples: Dict[int, dict] = manager.find_handler.get_dragen_demux_samples(
        demux_results=demux_results,
        sample_sheet=sample_sheet,
    )

    sample: NovaSeqSample
    for sample in sample_sheet.samples:
        stats_sample: Sample = get_or_create_sample(
            manager=manager, sample=sample, project_name_to_id=project_name_to_id
        )

        if not stats_sample:
            continue

        dragen_demux_sample: DragenDemuxSample = demux_samples[sample.lane][sample.sample_id]

        get_or_create_dragen_unaligned(
            manager=manager,
            dragen_demux_sample=dragen_demux_sample,
            sample_id=stats_sample.sample_id,
            demux_id=demux_id,
            lane=sample.lane,
        )


def _create_bcl2fastq_samples(
    manager: StatsAPI,
    demux_results: DemuxResults,
    project_name_to_id: Dict[str, int],
    demux_id: int,
    sample_sheet: SampleSheet,
):
    """Handles sample creation: creates sample objects and unaligned objects in their respective
    tables in cgstats for samples demultiplexed with bcl2fastq"""

    demux_samples: Dict[int, Dict[str, DemuxSample]] = manager.find_handler.get_demux_samples(
        conversion_stats=demux_results.conversion_stats,
        demux_stats_path=demux_results.demux_stats_path,
        sample_sheet=sample_sheet,
    )

    sample: NovaSeqSample
    for sample in sample_sheet.samples:
        stats_sample: Sample = get_or_create_sample(
            manager=manager, sample=sample, project_name_to_id=project_name_to_id
        )

        if not stats_sample:
            continue

        unaligned: Optional[
            Unaligned
        ] = manager.find_handler.get_unaligned_by_sample_id_demux_id_and_lane(
            sample_id=stats_sample.sample_id, demux_id=demux_id, lane=sample.lane
        )
        if not unaligned:
            demux_sample: DemuxSample = demux_samples[sample.lane][sample.sample_id]
            create_unaligned(
                manager=manager,
                demux_sample=demux_sample,
                sample_id=stats_sample.sample_id,
                demux_id=demux_id,
            )


def create_samples(
    manager: StatsAPI,
    demux_results: DemuxResults,
    project_name_to_id: Dict[str, int],
    demux_id: int,
) -> None:
    """dispatches sample object and unaligned object creation for samples based on the
    bcl-converter used in demultiplexing."""
    LOG.info(f"Creating samples for flow cell {demux_results.flow_cell.full_name}")
    sample_sheet: SampleSheet = demux_results.flow_cell.get_sample_sheet()

    create_samples_function = {
        "dragen": _create_dragen_samples,
        "bcl2fastq": _create_bcl2fastq_samples,
    }
    create_samples_function[demux_results.bcl_converter](
        manager, demux_results, project_name_to_id, demux_id, sample_sheet
    )


def create_novaseq_flowcell(manager: StatsAPI, demux_results: DemuxResults):
    """Add a novaseq flowcell to CG stats"""
    LOG.info("Adding flowcell information to cgstats")

    support_parameters: Supportparams = get_or_create_support_parameters(
        manager=manager, demux_results=demux_results
    )

    datasource: Datasource = get_or_create_datasource(
        manager=manager,
        demux_results=demux_results,
        support_parameters_id=support_parameters.supportparams_id,
    )

    flow_cell: Flowcell = get_or_create_flow_cell(manager=manager, demux_results=demux_results)

    demux: Demux = get_or_create_demux(
        manager=manager,
        demux_results=demux_results,
        flow_cell_id=flow_cell.flowcell_id,
        datasource_id=datasource.datasource_id,
    )

    project_name_to_id = create_projects(manager=manager, project_names=demux_results.projects)

    create_samples(
        manager=manager,
        demux_results=demux_results,
        project_name_to_id=project_name_to_id,
        demux_id=demux.demux_id,
    )

    manager.commit()


def get_or_create_dragen_unaligned(
    manager: StatsAPI,
    dragen_demux_sample: DragenDemuxSample,
    sample_id: int,
    demux_id: int,
    lane: int,
) -> Unaligned:
    unaligned = manager.find_handler.get_unaligned_by_sample_id_demux_id_and_lane(
        sample_id=sample_id, demux_id=demux_id, lane=lane
    )

    if not unaligned:
        unaligned = create_dragen_unaligned(
            manager=manager,
            demux_sample=dragen_demux_sample,
            sample_id=sample_id,
            demux_id=demux_id,
        )
    return unaligned


def get_or_create_support_parameters(
    manager: StatsAPI, demux_results: DemuxResults
) -> Supportparams:
    """Create support parameters for demux or retrieve them if they already exist."""
    document_path = str(demux_results.results_dir)

    support_parameters = manager.find_handler.get_support_parameters_by_document_path(
        document_path=document_path
    )

    if not support_parameters:
        support_parameters = create_support_parameters(manager, demux_results)
    else:
        LOG.info("Support parameters already exists")

    return support_parameters


def get_document_stats_path_from_demux(demux_results: DemuxResults) -> str:
    stats_path = {
        "bcl2fastq": demux_results.conversion_stats_path,
        "dragen": demux_results.demux_stats_path,
    }
    document_path: str = str(stats_path[demux_results.bcl_converter])
    return document_path


def get_or_create_datasource(
    manager: StatsAPI, demux_results: DemuxResults, support_parameters_id: int
) -> Datasource:
    """Create datasource for demux or retrieve it if it already exists."""

    document_path: str = get_document_stats_path_from_demux(demux_results=demux_results)

    datasource: Datasource = manager.find_handler.get_datasource_by_document_path(
        document_path=document_path
    )

    if not datasource:
        datasource = create_datasource(manager, demux_results, support_parameters_id)
    else:
        LOG.info("Data source already exists")

    return datasource


def get_or_create_flow_cell(manager: StatsAPI, demux_results: DemuxResults) -> Flowcell:
    flowcell = manager.find_handler.get_flow_cell_by_name(flow_cell_name=demux_results.flow_cell.id)

    if not flowcell:
        flowcell = create_flowcell(manager, demux_results)
    else:
        LOG.info("Flowcell already exists")

    return flowcell


def get_or_create_demux(
    manager: StatsAPI, demux_results: DemuxResults, flow_cell_id: int, datasource_id: int
) -> Demux:
    demux: Optional[Demux] = manager.find_handler.get_demux_by_flow_cell_id_and_base_mask(
        flowcell_id=flow_cell_id
    )
    if not demux:
        demux = create_demux(
            manager=manager,
            demux_results=demux_results,
            flow_cell_id=flow_cell_id,
            datasource_id=datasource_id,
        )
    else:
        LOG.info("Demux object already exists")

    return demux


def get_or_create_project(manager: StatsAPI, project_name: str) -> Project:
    project: Optional[Project] = manager.find_handler.get_project_by_name(project_name)

    if project:
        LOG.info(f"Project {project_name} already exists")
        return project

    return create_project(manager=manager, project_name=project_name)



import logging
import re
from typing import Iterable, List, Optional

import click
from cg.models.cg_config import CGConfig
from cg.store import Store
from tabulate import tabulate

from cg.store.models import Family, Flowcell, Customer, Sample

LOG = logging.getLogger(__name__)
ANALYSIS_HEADERS = ["Analysis Date", "Pipeline", "Version"]
FAMILY_HEADERS = ["Family", "Name", "Customer", "Priority", "Panels", "Action"]
FLOW_CELL_HEADERS = ["Flowcell", "Type", "Sequencer", "Date", "Archived?", "Status"]
LINK_HEADERS = ["Sample", "Mother", "Father"]
SAMPLE_HEADERS = ["Sample", "Name", "Customer", "Application", "State", "Priority", "External?"]


@click.group(invoke_without_command=True)
@click.option("-i", "--identifier", help="made a guess what type you are looking for")
@click.pass_context
def get(context: click.Context, identifier: Optional[str]):
    """Get information about records in the database."""
    if identifier:
        if re.match(r"^[A-Z]{3}[0-9]{4,5}[A-Z]{1}[1-9]{1,3}$", identifier):
            context.invoke(get_sample, sample_ids=[identifier])
        elif re.match(r"^[a-z]*$", identifier):
            # try case information
            context.invoke(get_case, case_ids=[identifier])
        elif re.match(r"^[HC][A-Z0-9]{8}$", identifier):
            # try flowcell information
            context.invoke(get_flow_cell, flow_cell_id=identifier)
        else:
            LOG.error(f"{identifier}: can't predict identifier")
            raise click.Abort


@get.command("sample")
@click.option("--cases/--no-cases", default=True, help="Display related cases")
@click.option("-hf", "--hide-flow-cell", is_flag=True, help="Hide related flow cells")
@click.argument("sample-ids", nargs=-1)
@click.pass_context
def get_sample(context: click.Context, cases: bool, hide_flow_cell: bool, sample_ids: List[str]):
    """Get information about a sample."""
    status_db: Store = context.obj.status_db
    for sample_id in sample_ids:
        LOG.debug("%s: get info about sample", sample_id)
        existing_sample: Sample = status_db.get_sample_by_internal_id(internal_id=sample_id)
        if existing_sample is None:
            LOG.warning(f"{sample_id}: sample doesn't exist")
            continue
        row = [
            existing_sample.internal_id,
            existing_sample.name,
            existing_sample.customer.internal_id,
            existing_sample.application_version.application.tag,
            existing_sample.state,
            existing_sample.priority_human,
            "Yes" if existing_sample.application_version.application.is_external else "No",
        ]
        click.echo(tabulate([row], headers=SAMPLE_HEADERS, tablefmt="psql"))
        if cases:
            case_ids: List[str] = [
                link_obj.family.internal_id for link_obj in existing_sample.links
            ]
            context.invoke(get_case, case_ids=case_ids, samples=False)
        if not hide_flow_cell:
            for sample_flow_cell in existing_sample.flowcells:
                LOG.debug(f"Get info on flow cell: {sample_flow_cell.name}")
                context.invoke(get_flow_cell, flow_cell_id=sample_flow_cell.name, samples=False)


@get.command("analysis")
@click.argument("case-id")
@click.pass_obj
def get_analysis(context: CGConfig, case_id: str):
    """Get information about case analysis."""
    status_db: Store = context.status_db
    case: Family = status_db.get_case_by_internal_id(internal_id=case_id)
    if case is None:
        LOG.error(f"{case_id}: case doesn't exist")
        raise click.Abort
    LOG.debug(f"{case.internal_id}: get info about case analysis")

    for case_analysis in case.analyses:
        row = [
            case_analysis.started_at,
            case_analysis.pipeline,
            case_analysis.pipeline_version,
        ]
        click.echo(tabulate([row], headers=ANALYSIS_HEADERS, tablefmt="psql"))


@get.command("relations")
@click.argument("case-id")
@click.pass_obj
def get_case_relations(context: CGConfig, case_id: str):
    """Get information about case relations."""
    status_db: Store = context.status_db
    case: Family = status_db.get_case_by_internal_id(internal_id=case_id)
    if case is None:
        LOG.error(f"{case_id}: case doesn't exist")
        raise click.Abort

    LOG.debug(f"{case.internal_id}: get info about case relations")

    for case_link in case.links:
        row = [
            case_link.sample.internal_id if case_link.sample else "",
            case_link.mother.internal_id if case_link.mother else "",
            case_link.father.internal_id if case_link.father else "",
        ]
        click.echo(tabulate([row], headers=LINK_HEADERS, tablefmt="psql"))


@get.command("case")
@click.option("-c", "--customer-id", help="internal id for customer to filter by")
@click.option("-n", "--name", is_flag=True, help="search family by name")
@click.option("--samples/--no-samples", default=True, help="display related samples")
@click.option("--relate/--no-relate", default=True, help="display relations to samples")
@click.option("--analyses/--no-analyses", default=True, help="display related analyses")
@click.argument("case-ids", nargs=-1)
@click.pass_context
def get_case(
    context: click.Context,
    customer_id: str,
    name: bool,
    samples: bool,
    relate: bool,
    analyses: bool,
    case_ids: List[str],
):
    """Get information about a case."""
    status_db: Store = context.obj.status_db
    status_db_cases: List[Family] = []
    if name:
        customer: Customer = status_db.get_customer_by_internal_id(customer_internal_id=customer_id)
        if not customer:
            LOG.error(f"{customer_id}: customer not found")
            raise click.Abort
        status_db_cases: Iterable[Family] = status_db.get_cases_by_customer_and_case_name_search(
            customer=customer, case_name_search=case_ids[-1]
        )
    else:
        for case_id in case_ids:
            existing_case: Family = status_db.get_case_by_internal_id(internal_id=case_id)
            if not existing_case:
                LOG.error(f"{case_id}: case doesn't exist")
                raise click.Abort
            status_db_cases.append(existing_case)

    for status_db_case in status_db_cases:
        LOG.debug(f"{status_db_case.internal_id}: get info about case")
        row: List[str] = [
            status_db_case.internal_id,
            status_db_case.name,
            status_db_case.customer.internal_id,
            status_db_case.priority_human,
            ", ".join(status_db_case.panels),
            status_db_case.action or "NA",
        ]
        click.echo(tabulate([row], headers=FAMILY_HEADERS, tablefmt="psql"))
        if relate:
            context.invoke(get_case_relations, case_id=status_db_case.internal_id)
        if samples:
            sample_ids: List[str] = [
                link_obj.sample.internal_id for link_obj in status_db_case.links
            ]
            context.invoke(get_sample, sample_ids=sample_ids, cases=False)
        if analyses:
            context.invoke(get_analysis, case_id=status_db_case.internal_id)


@get.command("flow-cell")
@click.option("--samples/--no-samples", default=True, help="Display related samples")
@click.argument("flow-cell-id")
@click.pass_context
def get_flow_cell(context: click.Context, samples: bool, flow_cell_id: str):
    """Get information about a flow cell and the samples on it."""
    status_db: Store = context.obj.status_db
    existing_flow_cell: Flowcell = status_db.get_flow_cell_by_name(flow_cell_name=flow_cell_id)
    if not existing_flow_cell:
        LOG.error(f"{flow_cell_id}: flow cell not found")
        raise click.Abort
    row: List[str] = [
        existing_flow_cell.name,
        existing_flow_cell.sequencer_type,
        existing_flow_cell.sequencer_name,
        existing_flow_cell.sequenced_at.date(),
        existing_flow_cell.archived_at.date() if existing_flow_cell.archived_at else "No",
        existing_flow_cell.status,
    ]
    click.echo(tabulate([row], headers=FLOW_CELL_HEADERS, tablefmt="psql"))
    if samples:
        sample_ids: List[str] = [
            sample_obj.internal_id for sample_obj in existing_flow_cell.samples
        ]
        if sample_ids:
            context.invoke(get_sample, sample_ids=sample_ids, cases=False)
        else:
            LOG.warning("No samples found on flow cell")



"""Tests for the BaseHandle class."""
from typing import Type
from mock import MagicMock

import pytest
from dataclasses import astuple

from sqlalchemy.orm import Query
from cg.constants.subject import PhenotypeStatus
from cg.store.api.base import BaseHandler
from cg.store.models import Model


@pytest.mark.parametrize("table", astuple(BaseHandler(MagicMock())))
def test__get_query(base_store, table: Type[Model]):
    """Tests the _get_query function for all attributes of BaseHandler ie tables in the database."""
    assert isinstance(base_store._get_query(table=table), Query)


def test_get_latest_analyses_for_cases_query(
    analysis_store, helpers, timestamp_now, timestamp_yesterday
):
    """Tests that analyses that are not latest are not returned."""

    # GIVEN an analysis a newer analysis exists
    case = helpers.add_case(analysis_store)
    analysis_oldest = helpers.add_analysis(
        analysis_store,
        case=case,
        started_at=timestamp_yesterday,
        uploaded_at=timestamp_yesterday,
        delivery_reported_at=None,
    )
    analysis_store.session.add(analysis_oldest)
    analysis_store.session.commit()
    analysis_newest = helpers.add_analysis(
        analysis_store,
        case=case,
        started_at=timestamp_now,
        uploaded_at=timestamp_now,
        delivery_reported_at=None,
    )
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp_now)
    analysis_store.relate_sample(
        family=analysis_oldest.family, sample=sample, status=PhenotypeStatus.UNKNOWN
    )

    # WHEN calling the analyses_to_delivery_report
    analyses: Query = analysis_store._get_latest_analyses_for_cases_query()

    # THEN analyses is a query
    assert isinstance(analyses, Query)

    # THEN only the newest analysis should be returned
    assert analysis_newest in analyses
    assert analysis_oldest not in analyses



from enum import Enum
from typing import List, Optional

from cg.constants import DataDelivery, Pipeline
from pydantic import BaseModel, constr, validator

from cg.store.models import Application, Family, Customer, Pool, Sample


class ControlEnum(str, Enum):
    not_control = ""
    positive = "positive"
    negative = "negative"


class SexEnum(str, Enum):
    male = "male"
    female = "female"
    unknown = "unknown"


class PriorityEnum(str, Enum):
    research = "research"
    standard = "standard"
    priority = "priority"
    express = "express"
    clinical_trials = "clinical_trials"


class ContainerEnum(str, Enum):
    no_container = "No container"
    plate = "96 well plate"
    tube = "Tube"


class StatusEnum(str, Enum):
    affected = "affected"
    unaffected = "unaffected"
    unknown = "unknown"


NAME_PATTERN = r"^[A-Za-z0-9-]*$"


class OrderSample(BaseModel):
    age_at_sampling: Optional[str]
    application: constr(max_length=Application.tag.property.columns[0].type.length)
    capture_kit: Optional[str]
    collection_date: Optional[str]
    comment: Optional[constr(max_length=Sample.comment.property.columns[0].type.length)]
    concentration: Optional[float]
    concentration_sample: Optional[float]
    container: Optional[ContainerEnum]
    container_name: Optional[str]
    control: Optional[str]
    customer: Optional[constr(max_length=Customer.internal_id.property.columns[0].type.length)]
    custom_index: Optional[str]
    data_analysis: Pipeline
    data_delivery: DataDelivery
    elution_buffer: Optional[str]
    extraction_method: Optional[str]
    family_name: Optional[
        constr(
            regex=NAME_PATTERN,
            min_length=2,
            max_length=Family.name.property.columns[0].type.length,
        )
    ]
    father: Optional[
        constr(regex=NAME_PATTERN, max_length=Sample.name.property.columns[0].type.length)
    ]
    formalin_fixation_time: Optional[int]
    index: Optional[str]
    index_number: Optional[str]
    index_sequence: Optional[str]
    internal_id: Optional[constr(max_length=Sample.internal_id.property.columns[0].type.length)]
    lab_code: Optional[str]
    mother: Optional[
        constr(regex=NAME_PATTERN, max_length=Sample.name.property.columns[0].type.length)
    ]
    name: constr(
        regex=NAME_PATTERN,
        min_length=2,
        max_length=Sample.name.property.columns[0].type.length,
    )
    organism: Optional[str]
    organism_other: Optional[str]
    original_lab: Optional[str]
    original_lab_address: Optional[str]
    phenotype_groups: Optional[List[str]]
    phenotype_terms: Optional[List[str]]
    pool: Optional[constr(max_length=Pool.name.property.columns[0].type.length)]
    post_formalin_fixation_time: Optional[int]
    pre_processing_method: Optional[str]
    priority: PriorityEnum = PriorityEnum.standard
    quantity: Optional[int]
    reagent_label: Optional[str]
    reference_genome: Optional[
        constr(max_length=Sample.reference_genome.property.columns[0].type.length)
    ]
    region: Optional[str]
    region_code: Optional[str]
    require_qc_ok: bool = False
    rml_plate_name: Optional[str]
    selection_criteria: Optional[str]
    sex: SexEnum = SexEnum.unknown
    source: Optional[str]
    status: StatusEnum = StatusEnum.unknown
    subject_id: Optional[
        constr(regex=NAME_PATTERN, max_length=Sample.subject_id.property.columns[0].type.length)
    ]
    tissue_block_size: Optional[str]
    tumour: bool = False
    tumour_purity: Optional[int]
    verified_organism: Optional[bool]
    volume: Optional[str]
    well_position: Optional[str]
    well_position_rml: Optional[str]

    @validator("priority", pre=True)
    def snake_case(cls, value: str):
        return value.lower().replace(" ", "_")



from cgmodels.cg.constants import Pipeline

from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.balsamic.analysis import BalsamicAnalysis
from cg.models.balsamic.config import BalsamicConfigJSON
from cg.models.balsamic.metrics import BalsamicTargetedQCMetrics


def test_instantiate_balsamic_analysis(cg_context, balsamic_config_raw, balsamic_metrics_raw):
    """Tests BALSAMIC analysis instance creation"""

    # GIVEN a config and metrics dictionaries and a BALSAMIC analysis API
    balsamic_analysis_api = BalsamicAnalysisAPI(cg_context, Pipeline.BALSAMIC)

    # WHEN instantiating a BALSAMIC analysis object
    balsamic_analysis = balsamic_analysis_api.parse_analysis(
        balsamic_config_raw, balsamic_metrics_raw
    )

    # THEN assert that it was successfully created
    assert isinstance(balsamic_analysis, BalsamicAnalysis)
    assert isinstance(balsamic_analysis.config, BalsamicConfigJSON)
    assert isinstance(balsamic_analysis.sample_metrics["ACC0000A1"], BalsamicTargetedQCMetrics)



from typing import List

import click
from cg.constants import CASE_ACTIONS, Pipeline
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Family, Sample
from ansi.colour import fg
from ansi.colour.fx import reset
from tabulate import tabulate

from cg.constants import Priority

from cg.utils.click.EnumChoice import EnumChoice

STATUS_OPTIONS = ["pending", "running", "completed", "failed", "error"]
CASE_HEADERS_LONG = [
    "Case",
    "Workflow",
    "Ordered",
    "Received",
    "Prepared",
    "Sequenced",
    "Flowcells",
    "Analysed",
    "Uploaded",
    "Delivered",
    "Delivery Reported",
    "Invoiced",
    "TAT",
]
ALWAYS_LONG_HEADERS = [
    CASE_HEADERS_LONG[0],
    CASE_HEADERS_LONG[1],
    CASE_HEADERS_LONG[6],
    CASE_HEADERS_LONG[7],
]
CASE_HEADERS_MEDIUM = []
CASE_HEADERS_SHORT = []

for header in CASE_HEADERS_LONG:
    if header not in ALWAYS_LONG_HEADERS:
        header = header[:3]

    CASE_HEADERS_MEDIUM.append(header)

    if header not in ALWAYS_LONG_HEADERS:
        header = header[0]

    CASE_HEADERS_SHORT.append(header)


@click.group()
def status():
    """View status of things."""
    pass


@status.command("analysis")
@click.pass_obj
def cases_to_be_analysed(context: CGConfig):
    """Which cases will be analysed?"""
    records: List[Family] = context.status_db.cases_to_analyze(pipeline=Pipeline.MIP_DNA)
    for case_obj in records:
        click.echo(case_obj)


def present_bool(a_dict, param, show_false=False):
    """presents boolean value in a human friendly format"""
    value = a_dict.get(param)

    if show_false:
        return (
            "-"
            if value is None
            else "✓"
            if value is True
            else "✗"
            if value is False
            else str(value)
        )

    return "-" if value is None else "✓" if value is True else "" if value is False else str(value)


def present_date(a_dict, param, show_negative, show_time):
    """presents datetime value in a human friendly format"""
    value = a_dict.get(param)

    if not show_time and value and value.date:
        value = value.date()

    if show_negative:
        return str(value)

    return "" if not value else value or str(value)


def present_string(a_dict, param, show_negative):
    """presents string value in a human friendly format"""
    value = a_dict.get(param)

    if show_negative:
        return str(value)

    return "" if not value else value or str(value)


@status.command("cases")
@click.pass_obj
@click.option(
    "-o",
    "--output",
    "output_type",
    type=click.Choice(["bool", "count", "date", "datetime"]),
    default="bool",
    help="how to display status",
)
@click.option("--verbose", is_flag=True, help="show status information otherwise left out")
@click.option("--days", default=31, help="days to go back")
@click.option("--internal-id", help="search by internal id")
@click.option("--name", help="search by name given by customer")
@click.option("--case-action", type=click.Choice(CASE_ACTIONS), help="filter by case action")
@click.option("--priority", type=EnumChoice(Priority, use_value=False), help="filter by priority")
@click.option("--data-analysis", help="filter on case data_analysis")
@click.option("--sample-id", help="filter by sample id")
@click.option("-c", "--customer-id", help="filter by customer")
@click.option("-C", "--exclude-customer-id", help="exclude customer")
@click.option("-r", "--only-received", is_flag=True, help="only completely received cases")
@click.option("-R", "--exclude-received", is_flag=True, help="exclude completely received cases")
@click.option("-p", "--only-prepared", is_flag=True, help="only completely prepared cases")
@click.option("-P", "--exclude-prepared", is_flag=True, help="exclude completely prepared cases")
@click.option("-s", "--only-sequenced", is_flag=True, help="only completely sequenced cases")
@click.option("-S", "--exclude-sequenced", is_flag=True, help="exclude completely sequenced cases")
@click.option("-a", "--only-analysed", is_flag=True, help="only analysed cases")
@click.option("-A", "--exclude-analysed", is_flag=True, help="exclude analysed cases")
@click.option("-u", "--only-uploaded", is_flag=True, help="only uploaded cases")
@click.option("-U", "--exclude-uploaded", is_flag=True, help="exclude uploaded cases")
@click.option("-d", "--only-delivered", is_flag=True, help="only LIMS delivered cases")
@click.option("-D", "--exclude-delivered", is_flag=True, help="exclude LIMS delivered cases")
@click.option("--dr", "--only-delivery-reported", is_flag=True, help="only delivery reported cases")
@click.option(
    "--DR", "--exclude-delivery-reported", is_flag=True, help="exclude delivery " "reported cases"
)
@click.option("-i", "--only-invoiced", is_flag=True, help="only completely invoiced cases")
@click.option("-I", "--exclude-invoiced", is_flag=True, help="exclude completely invoiced cases")
def status_of_cases(
    context: CGConfig,
    output_type,
    verbose,
    days,
    internal_id,
    name,
    case_action,
    priority,
    customer_id,
    data_analysis,
    sample_id,
    only_received,
    only_prepared,
    only_sequenced,
    only_analysed,
    only_uploaded,
    only_delivered,
    only_delivery_reported,
    only_invoiced,
    exclude_customer_id,
    exclude_received,
    exclude_prepared,
    exclude_sequenced,
    exclude_analysed,
    exclude_uploaded,
    exclude_delivered,
    exclude_delivery_reported,
    exclude_invoiced,
):
    """progress of each case"""
    status_db: Store = context.status_db
    records: List[Family] = status_db.cases(
        days=days,
        internal_id=internal_id,
        name=name,
        case_action=case_action,
        priority=priority,
        customer_id=customer_id,
        exclude_customer_id=exclude_customer_id,
        data_analysis=data_analysis,
        sample_id=sample_id,
        only_received=only_received,
        only_prepared=only_prepared,
        only_sequenced=only_sequenced,
        only_analysed=only_analysed,
        only_uploaded=only_uploaded,
        only_delivered=only_delivered,
        only_delivery_reported=only_delivery_reported,
        only_invoiced=only_invoiced,
        exclude_received=exclude_received,
        exclude_prepared=exclude_prepared,
        exclude_sequenced=exclude_sequenced,
        exclude_analysed=exclude_analysed,
        exclude_uploaded=exclude_uploaded,
        exclude_delivered=exclude_delivered,
        exclude_delivery_reported=exclude_delivery_reported,
        exclude_invoiced=exclude_invoiced,
    )
    case_rows = []

    if output_type == "bool":
        case_header = CASE_HEADERS_SHORT

    elif output_type == "count":
        case_header = CASE_HEADERS_MEDIUM

    elif output_type in ("date", "datetime"):
        case_header = CASE_HEADERS_LONG

    for case in records:
        tat_number = case.get("tat")
        max_tat = case.get("max_tat")

        if (
            case.get("samples_received_bool")
            and case.get("samples_delivered_bool")
            and tat_number <= max_tat
        ):
            tat_color = fg.green
        elif tat_number == max_tat:
            tat_color = fg.yellow
        elif tat_number > max_tat:
            tat_color = fg.red
        else:
            tat_color = fg.white

        color_start = tat_color
        color_end = str(reset)

        title = color_start + f"{case.get('internal_id')}"

        if name:
            title = f"{title} ({case.get('name')})"

        if (
            not case.get("case_external_bool")
            and case.get("samples_received_bool")
            and case.get("samples_delivered_bool")
        ):
            tat = f"{tat_number}/{max_tat}" + color_end
        elif case.get("case_external_bool") and case.get("analysis_uploaded_bool"):
            tat = f"{tat_number}/{max_tat}" + color_end
        else:
            tat = f"({tat_number})/{max_tat}" + color_end

        data_analysis = f"{case.get('data_analysis')}"

        show_time = output_type == "datetime"

        ordered = present_date(case, "ordered_at", verbose, show_time)

        if output_type == "bool":
            received = present_bool(case, "samples_received_bool", verbose)
            prepared = present_bool(case, "samples_prepared_bool", verbose)
            sequenced = present_bool(case, "samples_sequenced_bool", verbose)
            flowcell = present_bool(case, "flowcells_on_disk_bool", verbose)
            analysed_bool = present_bool(case, "analysis_completed_bool", verbose)

            if case.get("analysis_completed_at"):
                analysed = f"{analysed_bool}"
            elif case.get("analysis_status"):
                analysed = (
                    f"{analysed_bool}{present_string(case, 'analysis_status',verbose)}"
                    f" {case.get('analysis_completion')}%"
                )
            else:
                analysed = f"{analysed_bool}{present_string(case, 'case_action', verbose)}"

            uploaded = present_bool(case, "analysis_uploaded_bool", verbose)
            delivered = present_bool(case, "samples_delivered_bool", verbose)
            delivery_reported = present_bool(case, "analysis_delivery_reported_bool", verbose)
            invoiced = present_bool(case, "samples_invoiced_bool", verbose)

        elif output_type == "count":
            received = f"{case.get('samples_received')}/{case.get('samples_to_receive')}"
            prepared = f"{case.get('samples_prepared')}/{case.get('samples_to_prepare')}"
            sequenced = f"{case.get('samples_sequenced')}/{case.get('samples_to_sequence')}"
            flowcell = f"{case.get('flowcells_on_disk')}/{case.get('total_samples')}"

            if case.get("analysis_completed_at"):
                analysed = f"{present_date(case, 'analysis_completed_at', verbose, show_time)}"
            elif case.get("analysis_status"):
                analysed = (
                    f"{present_string(case, 'analysis_status', verbose)}"
                    f" {case.get('analysis_completion')}%"
                )
            else:
                analysed = f"{present_string(case, 'case_action', verbose)}"

            uploaded = present_date(case, "analysis_uploaded_at", verbose, show_time)
            delivered = f"{case.get('samples_delivered')}/{case.get('samples_to_deliver')}"
            delivery_reported = present_date(
                case, "analysis_delivery_reported_at", verbose, show_time
            )
            invoiced = f"{case.get('samples_invoiced')}/{case.get('samples_to_invoice')}"

        elif output_type in ("date", "datetime"):
            received = present_date(case, "samples_received_at", verbose, show_time)
            prepared = present_date(case, "samples_prepared_at", verbose, show_time)
            sequenced = present_date(case, "samples_sequenced_at", verbose, show_time)
            flowcell = present_string(case, "flowcells_status", verbose)

            if case.get("analysis_completed_at"):
                analysed = f"{present_date(case, 'analysis_completed_at', verbose, show_time)}"
            elif case.get("analysis_status"):
                analysed = (
                    f"{present_string(case, 'analysis_status', verbose)}"
                    f" {case.get('analysis_completion')}%"
                )
            else:
                analysed = f"{present_string(case, 'case_action', verbose)}"

            uploaded = present_date(case, "analysis_uploaded_at", verbose, show_time)
            delivered = present_date(case, "samples_delivered_at", verbose, show_time)
            delivery_reported = present_date(
                case, "analysis_delivery_reported_at", verbose, show_time
            )
            invoiced = present_date(case, "samples_invoiced_at", verbose, show_time)

        case_row = [
            title,
            data_analysis,
            ordered,
            received,
            prepared,
            sequenced,
            flowcell,
            analysed,
            uploaded,
            delivered,
            delivery_reported,
            invoiced,
            tat,
        ]
        case_rows.append(case_row)

    click.echo(tabulate(case_rows, headers=case_header, tablefmt="psql"))

    header_description = ""
    for i, _ in enumerate(case_header):
        if case_header[i] != CASE_HEADERS_LONG[i]:
            header_description = f"{header_description} {case_header[i]}={CASE_HEADERS_LONG[i]}"
    click.echo(header_description)


@status.command("samples")
@click.option("-s", "--skip", default=0, help="skip initial records")
@click.pass_obj
def status_of_samples(context: CGConfig, skip: int):
    """View status of samples."""
    status_db: Store = context.status_db
    records: List[Sample] = status_db.get_samples()[skip : skip + 30]
    for record in records:
        message = f"{record.internal_id} ({record.customer.internal_id})"
        if record.sequenced_at:
            color = "green"
            message += f" [SEQUENCED: {record.sequenced_at.date()}]"
        elif record.received_at and record.reads:
            color = "orange"
            message += f" [READS: {record.reads}]"
        elif record.received_at:
            color = "blue"
            message += f" [RECEIVED: {record.received_at.date()}]"
        else:
            color = "white"
            message += " [NOT RECEIVED]"
        click.echo(click.style(message, fg=color))


@status.command("families")
@click.option("-s", "--skip", default=0, help="skip initial records")
@click.pass_obj
def status_of_families(context: CGConfig, skip: int):
    """View status of families."""
    click.echo("red: prio > 1, blue: prio = 1, green: completed, yellow: action")
    status_db: Store = context.status_db
    records: List[Family] = status_db._get_query(table=Family).offset(skip).limit(30)
    for case_obj in records:
        color = "red" if case_obj.priority_int > 1 else "blue"
        message = f"{case_obj.internal_id} ({case_obj.priority_int})"
        if case_obj.analyses:
            message += f" {case_obj.analyses[0].completed_at.date()}"
            color = "green"
        if case_obj.action:
            message += f" [{case_obj.action.upper()}]"
            color = "yellow"
        click.echo(click.style(message, fg=color))



from flask_admin import Admin
from flask_cors import CORS
from flask_wtf.csrf import CSRFProtect
from cg.apps.lims import LimsAPI
from cg.apps.osticket import OsTicket
from cg.store.api.core import Store


class FlaskLims(LimsAPI):
    def __init__(self, app=None):
        if app:
            self.init_app(app)

    def init_app(self, app):
        config = {
            "lims": {
                "host": app.config["LIMS_HOST"],
                "username": app.config["LIMS_USERNAME"],
                "password": app.config["LIMS_PASSWORD"],
            }
        }
        super(FlaskLims, self).__init__(config)


class FlaskStore(Store):
    def __init__(self, app=None):
        if app:
            self.init_app(app)

    def init_app(self, app):
        uri = app.config["SQLALCHEMY_DATABASE_URI"]
        super(FlaskStore, self).__init__(uri)


cors = CORS(resources={r"/api/*": {"origins": "*"}}, supports_credentials=True)
csrf = CSRFProtect()
db = FlaskStore()

admin = Admin(name="Clinical Genomics")
lims = FlaskLims()
osticket = OsTicket()



"""Gene panel specific constants."""
from typing import List

from cg.utils.enums import StrEnum

GENOME_BUILD_37 = "37"
GENOME_BUILD_38 = "GRCh38"


class GenePanelMasterList(StrEnum):
    BRAIN: str = "BRAIN"
    CARDIOLOGY: str = "Cardiology"
    CILM: str = "CILM"
    CH: str = "CH"
    CTD: str = "CTD"
    DIAB: str = "DIAB"
    ENDO: str = "ENDO"
    EP: str = "EP"
    HEARING: str = "HEARING"
    HYDRO: str = "HYDRO"
    IBMFS: str = "IBMFS"
    IEM: str = "IEM"
    IF: str = "IF"
    MCARTA: str = "mcarta"
    MHT: str = "MHT"
    MIT: str = "MIT"
    MOVE: str = "MOVE"
    MTDNA: str = "mtDNA"
    NBS_M: str = "NBS-M"
    NEURODEG: str = "NEURODEG"
    NMD: str = "NMD"
    PEDHEP: str = "PEDHEP"
    PID: str = "PID"
    PIDCAD: str = "PIDCAD"
    RETINA: str = "RETINA"
    OMIM_AUTO: str = "OMIM-AUTO"
    PANELAPP_GREEN: str = "PANELAPP-GREEN"
    SKD: str = "SKD"
    SOVM: str = "SOVM"

    @classmethod
    def get_panel_names(cls, panels=None) -> List[str]:
        """Return requested panel names from Master list, or all panels if none are specified."""
        return (
            [panel_name.value for panel_name in panels]
            if panels
            else list(map(lambda panel: panel.value, cls))
        )
